import sys

import pytest

import salt.modules.pkg_resource as pkg_resource
import salt.modules.solarisipspkg as solarisips
import salt.utils.data
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


@pytest.mark.skipif(
    sys.platform != "solaris", reason="Skip when not running on Solaris"
)
class IpsTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.solarisips
    """

    def setup_loader_modules(self):
        self.opts = opts = salt.config.DEFAULT_MINION_OPTS.copy()
        utils = salt.loader.utils(opts, whitelist=["pkg", "path", "platform"])
        return {
            pkg_resource: {
                "__grains__": {
                    "osarch": "sparcv9",
                    "os_family": "Solaris",
                    "osmajorrelease": 11,
                    "kernelrelease": 5.11,
                },
            },
            solarisips: {"__opts__": opts, "__utils__": utils},
        }

    def test_install_single_package(self):
        """
        Test installing a single package
        """
        pkg_list_pre = {
            "pkg://solaris/compress/bzip2": (
                "1.0.6,5.11-0.175.3.10.0.4.0:20160630T215500Z"
            ),
            "pkg://solaris/compress/gzip": "1.5,5.11-0.175.3.0.0.30.0:20150821T161446Z",
            "pkg://solaris/compress/p7zip": (
                "16.2.3,5.11-0.175.3.34.0.2.0:20180614T204908Z"
            ),
        }
        pkg_list_post = {
            "pkg://solaris/compress/bzip2": (
                "1.0.6,5.11-0.175.3.10.0.4.0:20160630T215500Z"
            ),
            "pkg://solaris/compress/gzip": "1.5,5.11-0.175.3.0.0.30.0:20150821T161446Z",
            "pkg://solaris/compress/p7zip": (
                "16.2.3,5.11-0.175.3.34.0.2.0:20180614T204908Z"
            ),
            "pkg://solaris/text/less": "458,5.11-0.175.3.0.0.30.0:20150821T172730Z",
        }
        install_cmd = {
            "pid": 1234,
            "retcode": 0,
            "stderr": "",
            "stdout": "",
        }
        mock_install_cmd = MagicMock(return_value=install_cmd)
        list_pkgs_responses = [pkg_list_pre, pkg_list_post]
        with patch.object(solarisips, "is_installed", return_value=False), patch.object(
            solarisips, "list_pkgs", side_effect=list_pkgs_responses
        ), patch.dict(solarisips.__salt__, {"cmd.run_all": mock_install_cmd}):
            result = solarisips.install(name="less", refresh=False)
            self.assertEqual(
                result, salt.utils.data.compare_dicts(pkg_list_pre, pkg_list_post)
            )

    def test_install_list_pkgs(self):
        """
        Test installing a list of packages
        """
        pkg_list_pre = {
            "pkg://solaris/compress/bzip2": (
                "1.0.6,5.11-0.175.3.10.0.4.0:20160630T215500Z"
            ),
            "pkg://solaris/compress/gzip": "1.5,5.11-0.175.3.0.0.30.0:20150821T161446Z",
            "pkg://solaris/compress/p7zip": (
                "16.2.3,5.11-0.175.3.34.0.2.0:20180614T204908Z"
            ),
        }
        pkg_list_post = {
            "pkg://solaris/compress/bzip2": (
                "1.0.6,5.11-0.175.3.10.0.4.0:20160630T215500Z"
            ),
            "pkg://solaris/compress/gzip": "1.5,5.11-0.175.3.0.0.30.0:20150821T161446Z",
            "pkg://solaris/compress/p7zip": (
                "16.2.3,5.11-0.175.3.34.0.2.0:20180614T204908Z"
            ),
            "pkg://solaris/text/less": "458,5.11-0.175.3.0.0.30.0:20150821T172730Z",
            "pkg://solaris/system/library/security/libsasl": (
                "0.5.11,5.11-0.175.3.32.0.1.0:20180406T191209Z"
            ),
        }
        install_cmd = {
            "pid": 1234,
            "retcode": 0,
            "stderr": "",
            "stdout": "",
        }
        mock_install_cmd = MagicMock(return_value=install_cmd)
        list_pkgs_responses = [pkg_list_pre, pkg_list_post]
        with patch.object(solarisips, "is_installed", return_value=False), patch.object(
            solarisips, "list_pkgs", side_effect=list_pkgs_responses
        ), patch.dict(solarisips.__salt__, {"cmd.run_all": mock_install_cmd}):
            result = solarisips.install(pkgs=["less", "libsasl"], refresh=False)
            self.assertEqual(
                result, salt.utils.data.compare_dicts(pkg_list_pre, pkg_list_post)
            )

    def test_install_dict_pkgs_no_version(self):
        """
        Test installing a list of packages
        """
        pkg_list_pre = {
            "pkg://solaris/compress/bzip2": (
                "1.0.6,5.11-0.175.3.10.0.4.0:20160630T215500Z"
            ),
            "pkg://solaris/compress/gzip": "1.5,5.11-0.175.3.0.0.30.0:20150821T161446Z",
            "pkg://solaris/compress/p7zip": (
                "16.2.3,5.11-0.175.3.34.0.2.0:20180614T204908Z"
            ),
        }
        pkg_list_post = {
            "pkg://solaris/compress/bzip2": (
                "1.0.6,5.11-0.175.3.10.0.4.0:20160630T215500Z"
            ),
            "pkg://solaris/compress/gzip": "1.5,5.11-0.175.3.0.0.30.0:20150821T161446Z",
            "pkg://solaris/compress/p7zip": (
                "16.2.3,5.11-0.175.3.34.0.2.0:20180614T204908Z"
            ),
            "pkg://solaris/text/less": "458,5.11-0.175.3.0.0.30.0:20150821T172730Z",
            "pkg://solaris/system/library/security/libsasl": (
                "0.5.11,5.11-0.175.3.32.0.1.0:20180406T191209Z"
            ),
        }
        install_cmd = {
            "pid": 1234,
            "retcode": 0,
            "stderr": "",
            "stdout": "",
        }
        mock_install_cmd = MagicMock(return_value=install_cmd)
        list_pkgs_responses = [pkg_list_pre, pkg_list_post]
        with patch.object(solarisips, "is_installed", return_value=False), patch.object(
            solarisips, "list_pkgs", side_effect=list_pkgs_responses
        ), patch.dict(solarisips.__salt__, {"cmd.run_all": mock_install_cmd}):
            result = solarisips.install(
                pkgs=[{"less": ""}, {"libsasl": ""}], refresh=False
            )
            self.assertEqual(
                result, salt.utils.data.compare_dicts(pkg_list_pre, pkg_list_post)
            )

    def test_install_dict_pkgs_with_version(self):
        """
        Test installing a list of packages
        """
        pkg_list_pre = {
            "pkg://solaris/compress/bzip2": (
                "1.0.6,5.11-0.175.3.10.0.4.0:20160630T215500Z"
            ),
            "pkg://solaris/compress/gzip": "1.5,5.11-0.175.3.0.0.30.0:20150821T161446Z",
            "pkg://solaris/compress/p7zip": (
                "16.2.3,5.11-0.175.3.34.0.2.0:20180614T204908Z"
            ),
        }
        pkg_list_post = {
            "pkg://solaris/compress/bzip2": (
                "1.0.6,5.11-0.175.3.10.0.4.0:20160630T215500Z"
            ),
            "pkg://solaris/compress/gzip": "1.5,5.11-0.175.3.0.0.30.0:20150821T161446Z",
            "pkg://solaris/compress/p7zip": (
                "16.2.3,5.11-0.175.3.34.0.2.0:20180614T204908Z"
            ),
            "pkg://solaris/text/less": "458,5.11-0.175.3.0.0.30.0:20150821T172730Z",
            "pkg://solaris/system/library/security/libsasl": (
                "0.5.11,5.11-0.175.3.32.0.1.0:20180406T191209Z"
            ),
        }
        install_cmd = {
            "pid": 1234,
            "retcode": 0,
            "stderr": "",
            "stdout": "",
        }
        mock_install_cmd = MagicMock(return_value=install_cmd)
        list_pkgs_responses = [pkg_list_pre, pkg_list_post]
        with patch.object(solarisips, "is_installed", return_value=False), patch.object(
            solarisips, "list_pkgs", side_effect=list_pkgs_responses
        ), patch.dict(solarisips.__salt__, {"cmd.run_all": mock_install_cmd}):
            result = solarisips.install(
                pkgs=[
                    {"less": "458,5.11-0.175.3.0.0.30.0:20150821T172730Z"},
                    {"libsasl": "0.5.11,5.11-0.175.3.32.0.1.0:20180406T191209Z"},
                ],
                refresh=False,
            )
            self.assertEqual(
                result, salt.utils.data.compare_dicts(pkg_list_pre, pkg_list_post)
            )

    def test_install_already_installed_single_pkg(self):
        """
        Test installing a package that is already installed
        """
        result = None
        expected_result = {}
        with patch.object(solarisips, "is_installed", return_value=True):
            result = solarisips.install(name="less")
        self.assertEqual(result, expected_result)

    def test_install_dict_pkgs_with_version_validate_cmd(self):
        """
        Test installing a list of packages
        """

        def check_param(arg, **kwargs):
            self.assertEqual(
                arg,
                [
                    "pkg",
                    "install",
                    "-v",
                    "--accept",
                    "less@458,5.11-0.175.3.0.0.30.0:20150821T172730Z",
                    "libsasl@0.5.11,5.11-0.175.3.32.0.1.0:20180406T191209Z",
                ],
            )
            return {
                "pid": 1234,
                "retcode": 0,
                "stderr": "",
                "stdout": "",
            }

        pkg_list_pre = {
            "pkg://solaris/compress/bzip2": (
                "1.0.6,5.11-0.175.3.10.0.4.0:20160630T215500Z"
            ),
            "pkg://solaris/compress/gzip": "1.5,5.11-0.175.3.0.0.30.0:20150821T161446Z",
            "pkg://solaris/compress/p7zip": (
                "16.2.3,5.11-0.175.3.34.0.2.0:20180614T204908Z"
            ),
        }
        pkg_list_post = {
            "pkg://solaris/compress/bzip2": (
                "1.0.6,5.11-0.175.3.10.0.4.0:20160630T215500Z"
            ),
            "pkg://solaris/compress/gzip": "1.5,5.11-0.175.3.0.0.30.0:20150821T161446Z",
            "pkg://solaris/compress/p7zip": (
                "16.2.3,5.11-0.175.3.34.0.2.0:20180614T204908Z"
            ),
            "pkg://solaris/text/less": "458,5.11-0.175.3.0.0.30.0:20150821T172730Z",
            "pkg://solaris/system/library/security/libsasl": (
                "0.5.11,5.11-0.175.3.32.0.1.0:20180406T191209Z"
            ),
        }
        mock_install_cmd = MagicMock(side_effect=check_param)
        list_pkgs_responses = [pkg_list_pre, pkg_list_post]
        with patch.object(solarisips, "is_installed", return_value=False), patch.object(
            solarisips, "list_pkgs", side_effect=list_pkgs_responses
        ):
            with patch.dict(solarisips.__salt__, {"cmd.run_all": mock_install_cmd}):
                result = solarisips.install(
                    pkgs=[
                        {"less": "458,5.11-0.175.3.0.0.30.0:20150821T172730Z"},
                        {"libsasl": "0.5.11,5.11-0.175.3.32.0.1.0:20180406T191209Z"},
                    ],
                    refresh=False,
                )

    def test_install_dict_pkgs_no_version_validate_cmd(self):
        """
        Test installing a list of packages
        """

        def check_param(arg, **kwargs):
            self.assertEqual(
                arg, ["pkg", "install", "-v", "--accept", "less", "libsasl"]
            )
            return {
                "pid": 1234,
                "retcode": 0,
                "stderr": "",
                "stdout": "",
            }

        pkg_list_pre = {
            "pkg://solaris/compress/bzip2": (
                "1.0.6,5.11-0.175.3.10.0.4.0:20160630T215500Z"
            ),
            "pkg://solaris/compress/gzip": "1.5,5.11-0.175.3.0.0.30.0:20150821T161446Z",
            "pkg://solaris/compress/p7zip": (
                "16.2.3,5.11-0.175.3.34.0.2.0:20180614T204908Z"
            ),
        }
        pkg_list_post = {
            "pkg://solaris/compress/bzip2": (
                "1.0.6,5.11-0.175.3.10.0.4.0:20160630T215500Z"
            ),
            "pkg://solaris/compress/gzip": "1.5,5.11-0.175.3.0.0.30.0:20150821T161446Z",
            "pkg://solaris/compress/p7zip": (
                "16.2.3,5.11-0.175.3.34.0.2.0:20180614T204908Z"
            ),
            "pkg://solaris/text/less": "458,5.11-0.175.3.0.0.30.0:20150821T172730Z",
            "pkg://solaris/system/library/security/libsasl": (
                "0.5.11,5.11-0.175.3.32.0.1.0:20180406T191209Z"
            ),
        }
        mock_install_cmd = MagicMock(side_effect=check_param)
        list_pkgs_responses = [pkg_list_pre, pkg_list_post]
        with patch.object(solarisips, "is_installed", return_value=False), patch.object(
            solarisips, "list_pkgs", side_effect=list_pkgs_responses
        ):
            with patch.dict(solarisips.__salt__, {"cmd.run_all": mock_install_cmd}):
                result = solarisips.install(
                    pkgs=[{"less": ""}, {"libsasl": ""}], refresh=False
                )

    def test_install_list_pkgs_validate_cmd(self):
        """
        Test installing a list of packages
        """

        def check_param(arg, **kwargs):
            self.assertEqual(
                arg, ["pkg", "install", "-v", "--accept", "less", "libsasl"]
            )
            return {
                "pid": 1234,
                "retcode": 0,
                "stderr": "",
                "stdout": "",
            }

        pkg_list_pre = {
            "pkg://solaris/compress/bzip2": (
                "1.0.6,5.11-0.175.3.10.0.4.0:20160630T215500Z"
            ),
            "pkg://solaris/compress/gzip": "1.5,5.11-0.175.3.0.0.30.0:20150821T161446Z",
            "pkg://solaris/compress/p7zip": (
                "16.2.3,5.11-0.175.3.34.0.2.0:20180614T204908Z"
            ),
        }
        pkg_list_post = {
            "pkg://solaris/compress/bzip2": (
                "1.0.6,5.11-0.175.3.10.0.4.0:20160630T215500Z"
            ),
            "pkg://solaris/compress/gzip": "1.5,5.11-0.175.3.0.0.30.0:20150821T161446Z",
            "pkg://solaris/compress/p7zip": (
                "16.2.3,5.11-0.175.3.34.0.2.0:20180614T204908Z"
            ),
            "pkg://solaris/text/less": "458,5.11-0.175.3.0.0.30.0:20150821T172730Z",
            "pkg://solaris/system/library/security/libsasl": (
                "0.5.11,5.11-0.175.3.32.0.1.0:20180406T191209Z"
            ),
        }
        mock_install_cmd = MagicMock(side_effect=check_param)
        list_pkgs_responses = [pkg_list_pre, pkg_list_post]
        with patch.object(solarisips, "is_installed", return_value=False), patch.object(
            solarisips, "list_pkgs", side_effect=list_pkgs_responses
        ):
            with patch.dict(solarisips.__salt__, {"cmd.run_all": mock_install_cmd}):
                result = solarisips.install(pkgs=["less", "libsasl"], refresh=False)



"""
Connection module for Amazon SNS

:configuration: This module accepts explicit sns credentials but can also
    utilize IAM roles assigned to the instance through Instance Profiles. Dynamic
    credentials are then automatically obtained from AWS API and no further
    configuration is necessary. More Information available at:

    .. code-block:: text

        http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html

    If IAM roles are not used you need to specify them either in a pillar or
    in the minion's config file:

    .. code-block:: yaml

        sns.keyid: GKTADJGHEIQSXMKKRBJ08H
        sns.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

    A region may also be specified in the configuration:

    .. code-block:: yaml

        sns.region: us-east-1

    If a region is not specified, the default is us-east-1.

    It's also possible to specify key, keyid and region via a profile, either
    as a passed in dict, or as a string to pull from pillars or minion config:

    .. code-block:: yaml

        myprofile:
            keyid: GKTADJGHEIQSXMKKRBJ08H
            key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
            region: us-east-1

:depends: boto3
"""

# keep lint from choking on _get_conn and _cache_id
# pylint: disable=E0602


import logging

import salt.utils.versions

log = logging.getLogger(__name__)

# pylint: disable=unused-import
try:
    import boto3
    import botocore
    import jmespath

    logging.getLogger("boto3").setLevel(logging.CRITICAL)
    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False
# pylint: enable=unused-import


def __virtual__():
    """
    Only load if boto libraries exist.
    """
    has_boto_reqs = salt.utils.versions.check_boto_reqs()
    if has_boto_reqs is True:
        __utils__["boto3.assign_funcs"](__name__, "sns")
    return has_boto_reqs


def list_topics(region=None, key=None, keyid=None, profile=None):
    """
    Returns a list of the requester's topics

    CLI Example:

    .. code-block:: bash

        salt myminion boto3_sns.list_topics
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    res = {}
    NextToken = ""
    while NextToken is not None:
        ret = conn.list_topics(NextToken=NextToken)
        NextToken = ret.get("NextToken", None)
        arns = jmespath.search("Topics[*].TopicArn", ret)
        for t in arns:
            short_name = t.split(":")[-1]
            res[short_name] = t
    return res


def describe_topic(name, region=None, key=None, keyid=None, profile=None):
    """
    Returns details about a specific SNS topic, specified by name or ARN.

    CLI Example:

    .. code-block:: bash

        salt my_favorite_client boto3_sns.describe_topic a_sns_topic_of_my_choice
    """
    topics = list_topics(region=region, key=key, keyid=keyid, profile=profile)
    ret = {}
    for topic, arn in topics.items():
        if name in (topic, arn):
            ret = {"TopicArn": arn}
            ret["Subscriptions"] = list_subscriptions_by_topic(
                arn, region=region, key=key, keyid=keyid, profile=profile
            )
            ret["Attributes"] = get_topic_attributes(
                arn, region=region, key=key, keyid=keyid, profile=profile
            )
            # Grab extended attributes for the above subscriptions
            for sub in ret["Subscriptions"]:
                sub_arn = sub["SubscriptionArn"]
                if not sub_arn.startswith("arn:aws:sns:"):
                    # Sometimes a sub is in e.g. PendingAccept or other
                    # wierd states and doesn't have an ARN yet
                    log.debug("Subscription with invalid ARN %s skipped...", sub_arn)
                    continue
    return ret


def topic_exists(name, region=None, key=None, keyid=None, profile=None):
    """
    Check to see if an SNS topic exists.

    CLI Example:

    .. code-block:: bash

        salt myminion boto3_sns.topic_exists mytopic region=us-east-1
    """
    topics = list_topics(region=region, key=key, keyid=keyid, profile=profile)
    return name in list(topics.values()) + list(topics)


def create_topic(Name, region=None, key=None, keyid=None, profile=None):
    """
    Create an SNS topic.

    CLI Example:

    .. code-block:: bash

        salt myminion boto3_sns.create_topic mytopic region=us-east-1
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    try:
        ret = conn.create_topic(Name=Name)
        log.info("SNS topic %s created with ARN %s", Name, ret["TopicArn"])
        return ret["TopicArn"]
    except botocore.exceptions.ClientError as e:
        log.error("Failed to create SNS topic %s: %s", Name, e)
        return None
    except KeyError:
        log.error("Failed to create SNS topic %s", Name)
        return None


def delete_topic(TopicArn, region=None, key=None, keyid=None, profile=None):
    """
    Delete an SNS topic.

    CLI Example:

    .. code-block:: bash

        salt myminion boto3_sns.delete_topic mytopic region=us-east-1
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    try:
        conn.delete_topic(TopicArn=TopicArn)
        log.info("SNS topic %s deleted", TopicArn)
        return True
    except botocore.exceptions.ClientError as e:
        log.error("Failed to delete SNS topic %s: %s", name, e)
        return False


def get_topic_attributes(TopicArn, region=None, key=None, keyid=None, profile=None):
    """
    Returns all of the properties of a topic.  Topic properties returned might differ based on the
    authorization of the user.

    CLI Example:

    .. code-block:: bash

        salt myminion boto3_sns.get_topic_attributes someTopic region=us-west-1
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    try:
        return conn.get_topic_attributes(TopicArn=TopicArn).get("Attributes")
    except botocore.exceptions.ClientError as e:
        log.error("Failed to garner attributes for SNS topic %s: %s", TopicArn, e)
        return None


def set_topic_attributes(
    TopicArn,
    AttributeName,
    AttributeValue,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Set an attribute of a topic to a new value.

    CLI Example:

    .. code-block:: bash

        salt myminion boto3_sns.set_topic_attributes someTopic DisplayName myDisplayNameValue
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    try:
        conn.set_topic_attributes(
            TopicArn=TopicArn,
            AttributeName=AttributeName,
            AttributeValue=AttributeValue,
        )
        log.debug(
            "Set attribute %s=%s on SNS topic %s",
            AttributeName,
            AttributeValue,
            TopicArn,
        )
        return True
    except botocore.exceptions.ClientError as e:
        log.error(
            "Failed to set attribute %s=%s for SNS topic %s: %s",
            AttributeName,
            AttributeValue,
            TopicArn,
            e,
        )
        return False


def list_subscriptions_by_topic(
    TopicArn, region=None, key=None, keyid=None, profile=None
):
    """
    Returns a list of the subscriptions to a specific topic

    CLI Example:

    .. code-block:: bash

        salt myminion boto3_sns.list_subscriptions_by_topic mytopic region=us-east-1
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    NextToken = ""
    res = []
    try:
        while NextToken is not None:
            ret = conn.list_subscriptions_by_topic(
                TopicArn=TopicArn, NextToken=NextToken
            )
            NextToken = ret.get("NextToken", None)
            subs = ret.get("Subscriptions", [])
            res += subs
    except botocore.exceptions.ClientError as e:
        log.error("Failed to list subscriptions for SNS topic %s: %s", TopicArn, e)
        return None
    return res


def list_subscriptions(region=None, key=None, keyid=None, profile=None):
    """
    Returns a list of the requester's topics

    CLI Example:

    .. code-block:: bash

        salt myminion boto3_sns.list_subscriptions region=us-east-1
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    NextToken = ""
    res = []
    try:
        while NextToken is not None:
            ret = conn.list_subscriptions(NextToken=NextToken)
            NextToken = ret.get("NextToken", None)
            subs = ret.get("Subscriptions", [])
            res += subs
    except botocore.exceptions.ClientError as e:
        log.error("Failed to list SNS subscriptions: %s", e)
        return None
    return res


def get_subscription_attributes(
    SubscriptionArn, region=None, key=None, keyid=None, profile=None
):
    """
    Returns all of the properties of a subscription.

    CLI Example:

    .. code-block:: bash

        salt myminion boto3_sns.get_subscription_attributes somesubscription region=us-west-1
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    try:
        ret = conn.get_subscription_attributes(SubscriptionArn=SubscriptionArn)
        return ret["Attributes"]
    except botocore.exceptions.ClientError as e:
        log.error(
            "Failed to list attributes for SNS subscription %s: %s", SubscriptionArn, e
        )
        return None
    except KeyError:
        log.error("Failed to list attributes for SNS subscription %s", SubscriptionArn)
        return None


def set_subscription_attributes(
    SubscriptionArn,
    AttributeName,
    AttributeValue,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Set an attribute of a subscription to a new value.

    CLI Example:

    .. code-block:: bash

        salt myminion boto3_sns.set_subscription_attributes someSubscription RawMessageDelivery jsonStringValue
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    try:
        conn.set_subscription_attributes(
            SubscriptionArn=SubscriptionArn,
            AttributeName=AttributeName,
            AttributeValue=AttributeValue,
        )
        log.debug(
            "Set attribute %s=%s on SNS subscription %s",
            AttributeName,
            AttributeValue,
            SubscriptionArn,
        )
        return True
    except botocore.exceptions.ClientError as e:
        log.error(
            "Failed to set attribute %s=%s for SNS subscription %s: %s",
            AttributeName,
            AttributeValue,
            SubscriptionArn,
            e,
        )
        return False


def subscribe(
    TopicArn, Protocol, Endpoint, region=None, key=None, keyid=None, profile=None
):
    """
    Subscribe to a Topic.

    CLI Example:

    .. code-block:: bash

        salt myminion boto3_sns.subscribe mytopic https https://www.example.com/sns-endpoint
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    try:
        ret = conn.subscribe(TopicArn=TopicArn, Protocol=Protocol, Endpoint=Endpoint)
        log.info(
            "Subscribed %s %s to topic %s with SubscriptionArn %s",
            Protocol,
            Endpoint,
            TopicArn,
            ret["SubscriptionArn"],
        )
        return ret["SubscriptionArn"]
    except botocore.exceptions.ClientError as e:
        log.error("Failed to create subscription to SNS topic %s: %s", TopicArn, e)
        return None
    except KeyError:
        log.error("Failed to create subscription to SNS topic %s", TopicArn)
        return None


def unsubscribe(SubscriptionArn, region=None, key=None, keyid=None, profile=None):
    """
    Unsubscribe a specific SubscriptionArn of a topic.

    CLI Example:

    .. code-block:: bash

        salt myminion boto3_sns.unsubscribe my_subscription_arn region=us-east-1
    """
    if not SubscriptionArn.startswith("arn:aws:sns:"):
        # Grrr, AWS sent us an ARN that's NOT and ARN....
        # This can happen if, for instance, a subscription is left in PendingAcceptance or similar
        # Note that anything left in PendingConfirmation will be auto-deleted by AWS after 30 days
        # anyway, so this isn't as ugly a hack as it might seem at first...
        log.info(
            "Invalid subscription ARN `%s` passed - likely a PendingConfirmaton or"
            " such.  Skipping unsubscribe attempt as it would almost certainly fail...",
            SubscriptionArn,
        )
        return True
    subs = list_subscriptions(region=region, key=key, keyid=keyid, profile=profile)
    sub = [s for s in subs if s.get("SubscriptionArn") == SubscriptionArn]
    if not sub:
        log.error("Subscription ARN %s not found", SubscriptionArn)
        return False
    TopicArn = sub[0]["TopicArn"]
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    try:
        conn.unsubscribe(SubscriptionArn=SubscriptionArn)
        log.info("Deleted subscription %s from SNS topic %s", SubscriptionArn, TopicArn)
        return True
    except botocore.exceptions.ClientError as e:
        log.error("Failed to delete subscription %s: %s", SubscriptionArn, e)
        return False



"""
Test case for the vault SDB module
"""

import pytest

import salt.sdb.vault as vault
from tests.support.mock import MagicMock, call, patch


@pytest.fixture
def configure_loader_modules():
    return {
        vault: {
            "__opts__": {
                "vault": {
                    "url": "http://127.0.0.1",
                    "auth": {"token": "test", "method": "token"},
                }
            }
        }
    }


def test_set():
    """
    Test salt.sdb.vault.set function
    """
    version = {"v2": False, "data": None, "metadata": None, "type": None}
    mock_version = MagicMock(return_value=version)
    mock_vault = MagicMock()
    mock_vault.return_value.status_code = 200
    with patch.dict(vault.__utils__, {"vault.make_request": mock_vault}), patch.dict(
        vault.__utils__, {"vault.is_v2": mock_version}
    ):
        vault.set_("sdb://myvault/path/to/foo/bar", "super awesome")

    assert mock_vault.call_args_list == [
        call(
            "POST",
            "v1/sdb://myvault/path/to/foo",
            json={"bar": "super awesome"},
        )
    ]


def test_set_v2():
    """
    Test salt.sdb.vault.set function with kv v2 backend
    """
    version = {
        "v2": True,
        "data": "path/data/to/foo",
        "metadata": "path/metadata/to/foo",
        "type": "kv",
    }
    mock_version = MagicMock(return_value=version)
    mock_vault = MagicMock()
    mock_vault.return_value.status_code = 200
    with patch.dict(vault.__utils__, {"vault.make_request": mock_vault}), patch.dict(
        vault.__utils__, {"vault.is_v2": mock_version}
    ):
        vault.set_("sdb://myvault/path/to/foo/bar", "super awesome")

    assert mock_vault.call_args_list == [
        call(
            "POST",
            "v1/path/data/to/foo",
            json={"data": {"bar": "super awesome"}},
        )
    ]


def test_set_question_mark():
    """
    Test salt.sdb.vault.set_ while using the old
    deprecated solution with a question mark.
    """
    version = {"v2": False, "data": None, "metadata": None, "type": None}
    mock_version = MagicMock(return_value=version)
    mock_vault = MagicMock()
    mock_vault.return_value.status_code = 200
    with patch.dict(vault.__utils__, {"vault.make_request": mock_vault}), patch.dict(
        vault.__utils__, {"vault.is_v2": mock_version}
    ):
        vault.set_("sdb://myvault/path/to/foo?bar", "super awesome")

    assert mock_vault.call_args_list == [
        call(
            "POST",
            "v1/sdb://myvault/path/to/foo",
            json={"bar": "super awesome"},
        )
    ]


def test_get():
    """
    Test salt.sdb.vault.get function
    """
    version = {"v2": False, "data": None, "metadata": None, "type": None}
    mock_version = MagicMock(return_value=version)
    mock_vault = MagicMock()
    mock_vault.return_value.status_code = 200
    mock_vault.return_value.json.return_value = {"data": {"bar": "test"}}
    with patch.dict(vault.__utils__, {"vault.make_request": mock_vault}), patch.dict(
        vault.__utils__, {"vault.is_v2": mock_version}
    ):
        assert vault.get("sdb://myvault/path/to/foo/bar") == "test"

    assert mock_vault.call_args_list == [call("GET", "v1/sdb://myvault/path/to/foo")]


def test_get_v2():
    """
    Test salt.sdb.vault.get function with kv v2 backend
    """
    version = {
        "v2": True,
        "data": "path/data/to/foo",
        "metadata": "path/metadata/to/foo",
        "type": "kv",
    }
    mock_version = MagicMock(return_value=version)
    mock_vault = MagicMock()
    mock_vault.return_value.status_code = 200
    mock_vault.return_value.json.return_value = {"data": {"data": {"bar": "test"}}}
    with patch.dict(vault.__utils__, {"vault.make_request": mock_vault}), patch.dict(
        vault.__utils__, {"vault.is_v2": mock_version}
    ):
        assert vault.get("sdb://myvault/path/to/foo/bar") == "test"

    assert mock_vault.call_args_list == [call("GET", "v1/path/data/to/foo")]


def test_get_question_mark():
    """
    Test salt.sdb.vault.get while using the old
    deprecated solution with a question mark.
    """
    version = {"v2": False, "data": None, "metadata": None, "type": None}
    mock_version = MagicMock(return_value=version)
    mock_vault = MagicMock()
    mock_vault.return_value.status_code = 200
    mock_vault.return_value.json.return_value = {"data": {"bar": "test"}}
    with patch.dict(vault.__utils__, {"vault.make_request": mock_vault}), patch.dict(
        vault.__utils__, {"vault.is_v2": mock_version}
    ):
        assert vault.get("sdb://myvault/path/to/foo?bar") == "test"
    assert mock_vault.call_args_list == [call("GET", "v1/sdb://myvault/path/to/foo")]


def test_get_missing():
    """
    Test salt.sdb.vault.get function returns None
    if vault does not have an entry
    """
    version = {"v2": False, "data": None, "metadata": None, "type": None}
    mock_version = MagicMock(return_value=version)
    mock_vault = MagicMock()
    mock_vault.return_value.status_code = 404
    with patch.dict(vault.__utils__, {"vault.make_request": mock_vault}), patch.dict(
        vault.__utils__, {"vault.is_v2": mock_version}
    ):
        assert vault.get("sdb://myvault/path/to/foo/bar") is None

    assert mock_vault.call_args_list == [call("GET", "v1/sdb://myvault/path/to/foo")]


def test_get_missing_key():
    """
    Test salt.sdb.vault.get function returns None
    if vault does not have the key but does have the entry
    """
    version = {"v2": False, "data": None, "metadata": None, "type": None}
    mock_version = MagicMock(return_value=version)
    mock_vault = MagicMock()
    mock_vault.return_value.status_code = 200
    mock_vault.return_value.json.return_value = {"data": {"bar": "test"}}
    with patch.dict(vault.__utils__, {"vault.make_request": mock_vault}), patch.dict(
        vault.__utils__, {"vault.is_v2": mock_version}
    ):
        assert vault.get("sdb://myvault/path/to/foo/foo") is None

    assert mock_vault.call_args_list == [call("GET", "v1/sdb://myvault/path/to/foo")]



"""
This is the default list matcher.
"""

import logging

log = logging.getLogger(__name__)


def match(tgt, opts=None, minion_id=None):
    """
    Determines if this host is on the list
    """

    if not opts:
        opts = __opts__
    if not minion_id:
        minion_id = opts.get("id")

    try:
        if (
            f",{minion_id}," in tgt
            or tgt.startswith(minion_id + ",")
            or tgt.endswith("," + minion_id)
        ):
            return True
        # tgt is a string, which we know because the if statement above did not
        # cause one of the exceptions being caught. Therefore, look for an
        # exact match. (e.g. salt -L foo test.ping)
        return minion_id == tgt
    except (AttributeError, TypeError):
        # tgt is not a string, maybe it's a sequence type?
        try:
            return minion_id in tgt
        except Exception:  # pylint: disable=broad-except
            # tgt was likely some invalid type
            return False

    # We should never get here based on the return statements in the logic
    # above. If we do, it is because something above changed, and should be
    # considered as a bug. Log a warning to help us catch this.
    log.warning(
        "List matcher unexpectedly did not return, for target %s, "
        "this is probably a bug.",
        tgt,
    )
    return False



"""
Extract the pillar data for this minion
"""

import salt.pillar
import salt.utils.data
import salt.utils.dictupdate
from salt.defaults import DEFAULT_TARGET_DELIM

try:
    # Python 3
    from collections.abc import Mapping
except ImportError:
    # We still allow Py2 import because this could be executed in a machine with Py2.
    from collections import (  # pylint: disable=no-name-in-module,deprecated-class
        Mapping,
    )


def get(key, default="", merge=False, delimiter=DEFAULT_TARGET_DELIM):
    """
    .. versionadded:: 0.14.0

    Attempt to retrieve the named value from pillar, if the named value is not
    available return the passed default. The default return is an empty string.

    If the merge parameter is set to ``True``, the default will be recursively
    merged into the returned pillar data.

    The value can also represent a value in a nested dict using a ":" delimiter
    for the dict. This means that if a dict in pillar looks like this::

        {'pkg': {'apache': 'httpd'}}

    To retrieve the value associated with the apache key in the pkg dict this
    key can be passed::

        pkg:apache

    merge
        Specify whether or not the retrieved values should be recursively
        merged into the passed default.

        .. versionadded:: 2015.5.0

    delimiter
        Specify an alternate delimiter to use when traversing a nested dict

        .. versionadded:: 2015.5.0

    CLI Example:

    .. code-block:: bash

        salt '*' pillar.get pkg:apache
    """
    if merge:
        ret = salt.utils.data.traverse_dict_and_list(
            __pillar__.value(), key, {}, delimiter
        )
        if isinstance(ret, Mapping) and isinstance(default, Mapping):
            return salt.utils.dictupdate.update(default, ret)

    return salt.utils.data.traverse_dict_and_list(
        __pillar__.value(), key, default, delimiter
    )


def item(*args):
    """
    .. versionadded:: 0.16.2

    Return one or more pillar entries

    CLI Examples:

    .. code-block:: bash

        salt '*' pillar.item foo
        salt '*' pillar.item foo bar baz
    """
    ret = {}
    for arg in args:
        try:
            ret[arg] = __pillar__[arg]
        except KeyError:
            pass
    return ret


def raw(key=None):
    """
    Return the raw pillar data that is available in the module. This will
    show the pillar as it is loaded as the __pillar__ dict.

    CLI Example:

    .. code-block:: bash

        salt '*' pillar.raw

    With the optional key argument, you can select a subtree of the
    pillar raw data.::

        salt '*' pillar.raw key='roles'
    """
    if key:
        ret = __pillar__.get(key, {})
    else:
        ret = __pillar__.value()

    return ret


def keys(key, delimiter=DEFAULT_TARGET_DELIM):
    """
    .. versionadded:: 2015.8.0

    Attempt to retrieve a list of keys from the named value from the pillar.

    The value can also represent a value in a nested dict using a ":" delimiter
    for the dict, similar to how pillar.get works.

    delimiter
        Specify an alternate delimiter to use when traversing a nested dict

    CLI Example:

    .. code-block:: bash

        salt '*' pillar.keys web:sites
    """
    ret = salt.utils.data.traverse_dict_and_list(
        __pillar__.value(), key, KeyError, delimiter
    )

    if ret is KeyError:
        raise KeyError("Pillar key not found: {}".format(key))

    if not isinstance(ret, dict):
        raise ValueError("Pillar value in key {} is not a dict".format(key))

    return ret.keys()


def filter_by(lookup_dict, pillar, merge=None, default="default", base=None):
    """
    .. versionadded:: 2017.7.0

    Look up the given pillar in a given dictionary and return the result

    :param lookup_dict: A dictionary, keyed by a pillar, containing a value or
        values relevant to systems matching that pillar. For example, a key
        could be a pillar for a role and the value could the name of a package
        on that particular OS.

        The dictionary key can be a globbing pattern. The function will return
        the corresponding ``lookup_dict`` value where the pillar value matches
        the  pattern. For example:

        .. code-block:: bash

            # this will render 'got some salt' if ``role`` begins with 'salt'
            salt '*' pillar.filter_by '{salt*: got some salt, default: salt is not here}' role

    :param pillar: The name of a pillar to match with the system's pillar. For
        example, the value of the "role" pillar could be used to pull values
        from the ``lookup_dict`` dictionary.

        The pillar value can be a list. The function will return the
        ``lookup_dict`` value for a first found item in the list matching
        one of the ``lookup_dict`` keys.

    :param merge: A dictionary to merge with the results of the pillar
        selection from ``lookup_dict``. This allows another dictionary to
        override the values in the ``lookup_dict``.

    :param default: default lookup_dict's key used if the pillar does not exist
        or if the pillar value has no match on lookup_dict.  If unspecified
        the value is "default".

    :param base: A lookup_dict key to use for a base dictionary.  The
        pillar-selected ``lookup_dict`` is merged over this and then finally
        the ``merge`` dictionary is merged.  This allows common values for
        each case to be collected in the base and overridden by the pillar
        selection dictionary and the merge dictionary.  Default is unset.

    CLI Example:

    .. code-block:: bash

        salt '*' pillar.filter_by '{web: Serve it up, db: I query, default: x_x}' role
    """
    return salt.utils.data.filter_by(
        lookup_dict=lookup_dict,
        lookup=pillar,
        traverse=__pillar__.value(),
        merge=merge,
        default=default,
        base=base,
    )


# Allow pillar.data to also be used to return pillar data
items = raw
data = items



"""
Test cases for xml state
"""

import pytest

import salt.states.xml as xml
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {xml: {}}


def test_value_already_present():
    """
    Test for existing value_present
    """

    name = "testfile.xml"
    xpath = ".//list[@id='1']"
    value = "test value"

    state_return = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": f"{value} is already present",
    }

    with patch.dict(xml.__salt__, {"xml.get_value": MagicMock(return_value=value)}):
        assert xml.value_present(name, xpath, value) == state_return


def test_value_update():
    """
    Test for updating value_present
    """

    name = "testfile.xml"
    xpath = ".//list[@id='1']"
    value = "test value"

    old_value = "not test value"

    state_return = {
        "name": name,
        "changes": {name: {"new": value, "old": old_value}},
        "result": True,
        "comment": f"{name} updated",
    }

    with patch.dict(xml.__salt__, {"xml.get_value": MagicMock(return_value=old_value)}):
        with patch.dict(xml.__salt__, {"xml.set_value": MagicMock(return_value=True)}):
            assert xml.value_present(name, xpath, value) == state_return


def test_value_update_test():
    """
    Test for value_present test=True
    """

    name = "testfile.xml"
    xpath = ".//list[@id='1']"
    value = "test value"

    old_value = "not test value"

    state_return = {
        "name": name,
        "changes": {name: {"old": old_value, "new": value}},
        "result": None,
        "comment": f"{name} will be updated",
    }

    with patch.dict(xml.__salt__, {"xml.get_value": MagicMock(return_value=old_value)}):
        assert xml.value_present(name, xpath, value, test=True) == state_return


def test_value_update_invalid_xpath():
    """
    Test for value_present invalid xpath
    """

    name = "testfile.xml"
    xpath = ".//list[@id='1']"
    value = "test value"

    state_return = {
        "name": name,
        "changes": {},
        "result": False,
        "comment": f"xpath query {xpath} not found in {name}",
    }

    with patch.dict(xml.__salt__, {"xml.get_value": MagicMock(return_value=False)}):
        assert xml.value_present(name, xpath, value, test=True) == state_return



"""
    :codeauthor: :email:`Bo Maryniuk <bo@suse.de>`
"""

import copy
import os
import pathlib
import shutil
import sys
import tarfile
import tempfile
import types

import jinja2
import pytest

import salt.exceptions
import salt.utils.hashutils
import salt.utils.json
import salt.utils.platform
import salt.utils.stringutils

try:
    import backports  # noqa: F401
except ImportError:  # pragma: no cover
    fake_backports_dir = pathlib.Path(tempfile.mkdtemp(prefix="fake-backports"))
    fake_backports_dir.mkdir(parents=True, exist_ok=True)
    init_file = fake_backports_dir / "__init__.py"
    init_file.write_text("")
    backports = types.SimpleNamespace(__file__=str(init_file), __name__="backports")
    sys.modules["backports"] = backports
from salt.utils import thin
from salt.utils.stringutils import to_bytes as bts
from tests.support.helpers import TstSuiteLoggingHandler, VirtualEnv
from tests.support.mock import MagicMock, patch
from tests.support.runtests import RUNTIME_VARS


def patch_if(condition, *args, **kwargs):
    """
    Return a patch decorator if the provided condition is met
    """
    if condition:
        return patch(*args, **kwargs)

    def inner(func):
        return func

    return inner


class ThinTestContext:
    def __init__(self):
        self.jinja_fp = os.path.dirname(jinja2.__file__)

        self.ext_conf = {
            "test": {
                "py-version": [2, 7],
                "path": RUNTIME_VARS.SALT_CODE_DIR,
                "dependencies": {"jinja2": self.jinja_fp},
            }
        }
        self.tops = copy.deepcopy(self.ext_conf)
        self.tops["test"]["dependencies"] = [self.jinja_fp]
        self.tar = _tarfile(None).open()
        self.digest = salt.utils.hashutils.DigestCollector()
        self.exp_files = [
            os.path.join("salt", "payload.py"),
            os.path.join("jinja2", "__init__.py"),
        ]
        lib_root = os.path.join(RUNTIME_VARS.TMP, "fake-libs")
        self.fake_libs = {
            "distro": os.path.join(lib_root, "distro"),
            "jinja2": os.path.join(lib_root, "jinja2"),
            "yaml": os.path.join(lib_root, "yaml"),
            "tornado": os.path.join(lib_root, "tornado"),
            "msgpack": os.path.join(lib_root, "msgpack"),
        }

        code_dir = pathlib.Path(RUNTIME_VARS.CODE_DIR).resolve()
        self.exp_ret = {
            "distro": str(code_dir / "distro.py"),
            "jinja2": str(code_dir / "jinja2"),
            "yaml": str(code_dir / "yaml"),
            "tornado": str(code_dir / "tornado"),
            "msgpack": str(code_dir / "msgpack"),
            "certifi": str(code_dir / "certifi"),
            "singledispatch": str(code_dir / "singledispatch.py"),
            "looseversion": str(code_dir / "looseversion.py"),
            "packaging": str(code_dir / "packaging"),
            "backports": str(code_dir / "backports"),
        }
        self.exc_libs = ["jinja2", "yaml"]

    def cleanup(self):
        for lib, fp in self.fake_libs.items():
            if os.path.exists(fp):
                shutil.rmtree(fp)
        self.exc_libs = None
        self.jinja_fp = None
        self.ext_conf = None
        self.tops = None
        self.tar = None
        self.digest = None
        self.exp_files = None
        self.fake_libs = None
        self.exp_ret = None


@pytest.fixture
def thin_ctx():
    ctx = ThinTestContext()
    try:
        yield ctx
    finally:
        ctx.cleanup()


def _popen(return_value=None, side_effect=None, returncode=0):
    """
    Fake subprocess.Popen

    :return:
    """

    proc = MagicMock()
    proc.communicate = MagicMock(return_value=return_value, side_effect=side_effect)
    proc.returncode = returncode
    popen = MagicMock(return_value=proc)

    return popen


def _version_info(major=None, minor=None):
    """
    Fake version info.

    :param major:
    :param minor:
    :return:
    """

    class VersionInfo(tuple):
        pass

    vi = VersionInfo([major, minor])
    vi.major = major or sys.version_info.major
    vi.minor = minor or sys.version_info.minor

    return vi


def _tarfile(getinfo=False):
    """
    Fake tarfile handler.

    :return:
    """
    spec = ["add", "close"]
    if getinfo:
        spec.append("getinfo")

    tf = MagicMock()
    tf.open = MagicMock(return_value=MagicMock(spec=spec))

    return tf


@patch("salt.utils.thin.log", MagicMock())
@patch("salt.utils.thin.os.path.isfile", MagicMock(return_value=False))
def test_get_ext_tops_cfg_missing_dependencies(thin_ctx):
    """
    Test thin.get_ext_tops contains all required dependencies.

    :return:
    """
    cfg = {"namespace": {"py-version": [0, 0], "path": "/foo", "dependencies": []}}

    with pytest.raises(salt.exceptions.SaltSystemExit) as err:
        thin.get_ext_tops(cfg)
    assert "Missing dependencies" in str(err.value)
    assert thin.log.error.called
    assert "Missing dependencies" in thin.log.error.call_args[0][0]
    assert "jinja2, yaml, tornado, msgpack" in thin.log.error.call_args[0][0]


@patch("salt.exceptions.SaltSystemExit", Exception)
@patch("salt.utils.thin.log", MagicMock())
@patch("salt.utils.thin.os.path.isfile", MagicMock(return_value=False))
def test_get_ext_tops_cfg_missing_interpreter(thin_ctx):
    """
    Test thin.get_ext_tops contains interpreter configuration.

    :return:
    """
    cfg = {"namespace": {"path": "/foo", "dependencies": []}}
    with pytest.raises(salt.exceptions.SaltSystemExit) as err:
        thin.get_ext_tops(cfg)
    assert "missing specific locked Python version" in str(err.value)


@patch("salt.exceptions.SaltSystemExit", Exception)
@patch("salt.utils.thin.log", MagicMock())
@patch("salt.utils.thin.os.path.isfile", MagicMock(return_value=False))
def test_get_ext_tops_cfg_wrong_interpreter(thin_ctx):
    """
    Test thin.get_ext_tops contains correct interpreter configuration.

    :return:
    """
    cfg = {"namespace": {"path": "/foo", "py-version": 2, "dependencies": []}}

    with pytest.raises(salt.exceptions.SaltSystemExit) as err:
        thin.get_ext_tops(cfg)
    assert (
        "specific locked Python version should be a list of major/minor version"
        in str(err.value)
    )


@patch("salt.exceptions.SaltSystemExit", Exception)
@patch("salt.utils.thin.log", MagicMock())
@patch("salt.utils.thin.os.path.isfile", MagicMock(return_value=False))
def test_get_ext_tops_cfg_interpreter(thin_ctx):
    """
    Test thin.get_ext_tops interpreter configuration.

    :return:
    """
    cfg = {
        "namespace": {
            "path": "/foo",
            "py-version": [2, 6],
            "dependencies": {
                "jinja2": "",
                "yaml": "",
                "tornado": "",
                "msgpack": "",
            },
        }
    }

    with pytest.raises(salt.exceptions.SaltSystemExit):
        thin.get_ext_tops(cfg)
    assert len(thin.log.warning.mock_calls) == 4
    assert sorted(x[1][1] for x in thin.log.warning.mock_calls) == [
        "jinja2",
        "msgpack",
        "tornado",
        "yaml",
    ]
    assert (
        "Module test has missing configuration"
        == thin.log.warning.mock_calls[0][1][0] % "test"
    )


@patch("salt.utils.thin.log", MagicMock())
@patch("salt.utils.thin.os.path.isfile", MagicMock(return_value=False))
def test_get_ext_tops_dependency_config_check(thin_ctx):
    """
    Test thin.get_ext_tops dependencies are importable

    :return:
    """
    cfg = {
        "namespace": {
            "path": "/foo",
            "py-version": [2, 6],
            "dependencies": {
                "jinja2": "/jinja/foo.py",
                "yaml": "/yaml/",
                "tornado": "/tornado/wrong.rb",
                "msgpack": "msgpack.sh",
            },
        }
    }

    with pytest.raises(salt.exceptions.SaltSystemExit) as err:
        thin.get_ext_tops(cfg)

    assert (
        "Missing dependencies for the alternative version in the "
        "external configuration"
    ) in str(err.value)

    messages = {}
    for cl in thin.log.warning.mock_calls:
        messages[cl[1][1]] = cl[1][0] % (cl[1][1], cl[1][2])
    for mod in ["tornado", "yaml", "msgpack"]:
        assert "not a Python importable module" in messages[mod]
    assert "configured with not a file or does not exist" in messages["jinja2"]


@patch("salt.exceptions.SaltSystemExit", Exception)
@patch("salt.utils.thin.log", MagicMock())
@patch("salt.utils.thin.os.path.isfile", MagicMock(return_value=True))
def test_get_ext_tops_config_pass(thin_ctx):
    """
    Test thin.get_ext_tops configuration

    :return:
    """
    cfg = {
        "namespace": {
            "path": "/foo",
            "py-version": [2, 6],
            "dependencies": {
                "jinja2": "/jinja/foo.py",
                "yaml": "/yaml/",
                "tornado": "/tornado/tornado.py",
                "msgpack": "msgpack.py",
                "distro": "distro.py",
            },
        }
    }
    out = thin.get_ext_tops(cfg)
    assert out["namespace"]["py-version"] == cfg["namespace"]["py-version"]
    assert out["namespace"]["path"] == cfg["namespace"]["path"]
    assert sorted(out["namespace"]["dependencies"]) == sorted(
        [
            "/tornado/tornado.py",
            "/jinja/foo.py",
            "/yaml/",
            "msgpack.py",
            "distro.py",
        ]
    )


@patch("salt.utils.thin.sys.argv", [None, '{"foo": "bar"}'])
@patch("salt.utils.thin.get_tops", lambda **kw: kw)
def test_gte(thin_ctx):
    """
    Test thin.gte external call for processing the info about tops per interpreter.

    :return:
    """
    assert salt.utils.json.loads(thin.gte()).get("foo") == "bar"


def test_add_dep_path(thin_ctx):
    """
    Test thin._add_dependency function to setup dependency paths
    :return:
    """
    container = []
    for pth in ["/foo/bar.py", "/something/else/__init__.py"]:
        thin._add_dependency(container, type("obj", (), {"__file__": pth})())
    assert "__init__" not in container[1]
    assert container == ["/foo/bar.py", "/something/else"]


def test_thin_path(thin_ctx):
    """
    Test thin.thin_path returns the expected path.

    :return:
    """
    path = os.sep + os.path.join("path", "to")
    expected = os.path.join(path, "thin", "thin.tgz")
    assert thin.thin_path(path) == expected


def test_get_salt_call_script(thin_ctx):
    """
    Test get salt-call script rendered.

    :return:
    """
    out = thin._get_salt_call("foo", "bar", py26=[2, 6], py27=[2, 7], py34=[3, 4])
    for line in salt.utils.stringutils.to_str(out).split(os.linesep):
        if line.startswith("namespaces = {"):
            data = salt.utils.json.loads(line.replace("namespaces = ", "").strip())
            assert data.get("py26") == [2, 6]
            assert data.get("py27") == [2, 7]
            assert data.get("py34") == [3, 4]
        if line.startswith("syspaths = "):
            data = salt.utils.json.loads(line.replace("syspaths = ", ""))
            assert data == ["foo", "bar"]


def test_get_ext_namespaces_empty(thin_ctx):
    """
    Test thin._get_ext_namespaces function returns an empty dictionary on nothing
    :return:
    """
    for obj in [None, {}, []]:
        assert thin._get_ext_namespaces(obj) == {}


def test_get_ext_namespaces(thin_ctx):
    """
    Test thin._get_ext_namespaces function returns namespaces properly out of the config.
    :return:
    """
    cfg = {"ns": {"py-version": [2, 7]}}
    assert thin._get_ext_namespaces(cfg).get("ns") == (
        2,
        7,
    )
    assert isinstance(thin._get_ext_namespaces(cfg).get("ns"), tuple)


def test_get_ext_namespaces_failure(thin_ctx):
    """
    Test thin._get_ext_namespaces function raises an exception
    if python major/minor version is not configured.
    :return:
    """
    with pytest.raises(salt.exceptions.SaltSystemExit):
        thin._get_ext_namespaces({"ns": {}})


@patch(
    "salt.utils.thin.distro",
    type("distro", (), {"__file__": "/site-packages/distro"}),
)
@patch(
    "salt.utils.thin.salt",
    type("salt", (), {"__file__": "/site-packages/salt"}),
)
@patch(
    "salt.utils.thin.jinja2",
    type("jinja2", (), {"__file__": "/site-packages/jinja2"}),
)
@patch(
    "salt.utils.thin.yaml",
    type("yaml", (), {"__file__": "/site-packages/yaml"}),
)
@patch(
    "salt.utils.thin.tornado",
    type("tornado", (), {"__file__": "/site-packages/tornado"}),
)
@patch(
    "salt.utils.thin.msgpack",
    type("msgpack", (), {"__file__": "/site-packages/msgpack"}),
)
@patch(
    "salt.utils.thin.certifi",
    type("certifi", (), {"__file__": "/site-packages/certifi"}),
)
@patch(
    "salt.utils.thin.singledispatch",
    type("singledispatch", (), {"__file__": "/site-packages/sdp"}),
)
@patch(
    "salt.utils.thin.singledispatch_helpers",
    type("singledispatch_helpers", (), {"__file__": "/site-packages/sdp_hlp"}),
)
@patch(
    "salt.utils.thin.ssl_match_hostname",
    type("ssl_match_hostname", (), {"__file__": "/site-packages/ssl_mh"}),
)
@patch(
    "salt.utils.thin.markupsafe",
    type("markupsafe", (), {"__file__": "/site-packages/markupsafe"}),
)
@patch(
    "salt.utils.thin.backports",
    type("backports", (), {"__file__": "/site-packages/backports"}),
)
@patch(
    "salt.utils.thin.backports_abc",
    type("backports_abc", (), {"__file__": "/site-packages/backports_abc"}),
)
@patch(
    "salt.utils.thin.concurrent",
    type("concurrent", (), {"__file__": "/site-packages/concurrent"}),
)
@patch(
    "salt.utils.thin.py_contextvars",
    type("contextvars", (), {"__file__": "/site-packages/contextvars"}),
)
@patch(
    "salt.utils.thin.packaging",
    type("packaging", (), {"__file__": "/site-packages/packaging"}),
)
@patch(
    "salt.utils.thin.looseversion",
    type("looseversion", (), {"__file__": "/site-packages/looseversion"}),
)
@patch_if(
    salt.utils.thin.has_immutables,
    "salt.utils.thin.immutables",
    type("immutables", (), {"__file__": "/site-packages/immutables"}),
)
@patch("salt.utils.thin.log", MagicMock())
def test_get_tops(thin_ctx):
    """
    Test thin.get_tops to get top directories, based on the interpreter.
    :return:
    """
    base_tops = [
        "distro",
        "salt",
        "jinja2",
        "yaml",
        "tornado",
        "msgpack",
        "certifi",
        "sdp",
        "sdp_hlp",
        "ssl_mh",
        "markupsafe",
        "backports",
        "backports_abc",
        "concurrent",
        "contextvars",
        "looseversion",
        "packaging",
    ]
    if salt.utils.thin.has_immutables:
        base_tops.extend(["immutables"])
    tops = []
    for top in thin.get_tops(extra_mods="foo,bar"):
        if top.find("/") != -1:
            spl = "/"
        else:
            spl = os.sep
        tops.append(top.rsplit(spl, 1)[-1])
    assert len(tops) == len(base_tops)
    assert sorted(tops) == sorted(base_tops), sorted(tops)


@patch(
    "salt.utils.thin.distro",
    type("distro", (), {"__file__": "/site-packages/distro"}),
)
@patch(
    "salt.utils.thin.salt",
    type("salt", (), {"__file__": "/site-packages/salt"}),
)
@patch(
    "salt.utils.thin.jinja2",
    type("jinja2", (), {"__file__": "/site-packages/jinja2"}),
)
@patch(
    "salt.utils.thin.yaml",
    type("yaml", (), {"__file__": "/site-packages/yaml"}),
)
@patch(
    "salt.utils.thin.tornado",
    type("tornado", (), {"__file__": "/site-packages/tornado"}),
)
@patch(
    "salt.utils.thin.msgpack",
    type("msgpack", (), {"__file__": "/site-packages/msgpack"}),
)
@patch(
    "salt.utils.thin.certifi",
    type("certifi", (), {"__file__": "/site-packages/certifi"}),
)
@patch(
    "salt.utils.thin.singledispatch",
    type("singledispatch", (), {"__file__": "/site-packages/sdp"}),
)
@patch(
    "salt.utils.thin.singledispatch_helpers",
    type("singledispatch_helpers", (), {"__file__": "/site-packages/sdp_hlp"}),
)
@patch(
    "salt.utils.thin.ssl_match_hostname",
    type("ssl_match_hostname", (), {"__file__": "/site-packages/ssl_mh"}),
)
@patch(
    "salt.utils.thin.markupsafe",
    type("markupsafe", (), {"__file__": "/site-packages/markupsafe"}),
)
@patch(
    "salt.utils.thin.backports",
    type("backports", (), {"__file__": "/site-packages/backports"}),
)
@patch(
    "salt.utils.thin.backports_abc",
    type("backports_abc", (), {"__file__": "/site-packages/backports_abc"}),
)
@patch(
    "salt.utils.thin.concurrent",
    type("concurrent", (), {"__file__": "/site-packages/concurrent"}),
)
@patch(
    "salt.utils.thin.py_contextvars",
    type("contextvars", (), {"__file__": "/site-packages/contextvars"}),
)
@patch(
    "salt.utils.thin.packaging",
    type("packaging", (), {"__file__": "/site-packages/packaging"}),
)
@patch(
    "salt.utils.thin.looseversion",
    type("looseversion", (), {"__file__": "/site-packages/looseversion"}),
)
@patch_if(
    salt.utils.thin.has_immutables,
    "salt.utils.thin.immutables",
    type("immutables", (), {"__file__": "/site-packages/immutables"}),
)
@patch("salt.utils.thin.log", MagicMock())
def test_get_tops_extra_mods(thin_ctx):
    """
    Test thin.get_tops to get extra-modules alongside the top directories, based on the interpreter.
    :return:
    """
    base_tops = [
        "distro",
        "salt",
        "jinja2",
        "yaml",
        "tornado",
        "msgpack",
        "certifi",
        "sdp",
        "sdp_hlp",
        "ssl_mh",
        "concurrent",
        "markupsafe",
        "backports",
        "backports_abc",
        "contextvars",
        "looseversion",
        "packaging",
        "foo",
        "bar.py",
    ]
    if salt.utils.thin.has_immutables:
        base_tops.extend(["immutables"])
    libs = salt.utils.thin.find_site_modules("contextvars")
    foo = {"__file__": os.sep + os.path.join("custom", "foo", "__init__.py")}
    bar = {"__file__": os.sep + os.path.join("custom", "bar")}
    with patch("salt.utils.thin.find_site_modules", MagicMock(side_effect=[libs])):
        with patch(
            "builtins.__import__",
            MagicMock(side_effect=[type("foo", (), foo), type("bar", (), bar)]),
        ):
            tops = []
            for top in thin.get_tops(extra_mods="foo,bar"):
                if top.find("/") != -1:
                    spl = "/"
                else:
                    spl = os.sep
                tops.append(top.rsplit(spl, 1)[-1])
    assert len(tops) == len(base_tops)
    assert sorted(tops) == sorted(base_tops)


@patch(
    "salt.utils.thin.distro",
    type("distro", (), {"__file__": "/site-packages/distro"}),
)
@patch(
    "salt.utils.thin.salt",
    type("salt", (), {"__file__": "/site-packages/salt"}),
)
@patch(
    "salt.utils.thin.jinja2",
    type("jinja2", (), {"__file__": "/site-packages/jinja2"}),
)
@patch(
    "salt.utils.thin.yaml",
    type("yaml", (), {"__file__": "/site-packages/yaml"}),
)
@patch(
    "salt.utils.thin.tornado",
    type("tornado", (), {"__file__": "/site-packages/tornado"}),
)
@patch(
    "salt.utils.thin.msgpack",
    type("msgpack", (), {"__file__": "/site-packages/msgpack"}),
)
@patch(
    "salt.utils.thin.certifi",
    type("certifi", (), {"__file__": "/site-packages/certifi"}),
)
@patch(
    "salt.utils.thin.singledispatch",
    type("singledispatch", (), {"__file__": "/site-packages/sdp"}),
)
@patch(
    "salt.utils.thin.singledispatch_helpers",
    type("singledispatch_helpers", (), {"__file__": "/site-packages/sdp_hlp"}),
)
@patch(
    "salt.utils.thin.ssl_match_hostname",
    type("ssl_match_hostname", (), {"__file__": "/site-packages/ssl_mh"}),
)
@patch(
    "salt.utils.thin.markupsafe",
    type("markupsafe", (), {"__file__": "/site-packages/markupsafe"}),
)
@patch(
    "salt.utils.thin.backports",
    type("backports", (), {"__file__": "/site-packages/backports"}),
)
@patch(
    "salt.utils.thin.backports_abc",
    type("backports_abc", (), {"__file__": "/site-packages/backports_abc"}),
)
@patch(
    "salt.utils.thin.concurrent",
    type("concurrent", (), {"__file__": "/site-packages/concurrent"}),
)
@patch(
    "salt.utils.thin.py_contextvars",
    type("contextvars", (), {"__file__": "/site-packages/contextvars"}),
)
@patch(
    "salt.utils.thin.packaging",
    type("packaging", (), {"__file__": "/site-packages/packaging"}),
)
@patch(
    "salt.utils.thin.looseversion",
    type("looseversion", (), {"__file__": "/site-packages/looseversion"}),
)
@patch_if(
    salt.utils.thin.has_immutables,
    "salt.utils.thin.immutables",
    type("immutables", (), {"__file__": "/site-packages/immutables"}),
)
@patch("salt.utils.thin.log", MagicMock())
def test_get_tops_so_mods(thin_ctx):
    """
    Test thin.get_tops to get extra-modules alongside the top directories, based on the interpreter.
    :return:
    """
    base_tops = [
        "distro",
        "salt",
        "jinja2",
        "yaml",
        "tornado",
        "msgpack",
        "certifi",
        "sdp",
        "sdp_hlp",
        "ssl_mh",
        "concurrent",
        "markupsafe",
        "backports",
        "backports_abc",
        "contextvars",
        "looseversion",
        "packaging",
        "foo.so",
        "bar.so",
    ]
    if salt.utils.thin.has_immutables:
        base_tops.extend(["immutables"])
    libs = salt.utils.thin.find_site_modules("contextvars")
    with patch("salt.utils.thin.find_site_modules", MagicMock(side_effect=[libs])):
        with patch(
            "builtins.__import__",
            MagicMock(
                side_effect=[
                    type("salt", (), {"__file__": "/custom/foo.so"}),
                    type("salt", (), {"__file__": "/custom/bar.so"}),
                ]
            ),
        ):
            tops = []
            for top in thin.get_tops(so_mods="foo,bar"):
                if top.find("/") != -1:
                    spl = "/"
                else:
                    spl = os.sep
                tops.append(top.rsplit(spl, 1)[-1])
    assert len(tops) == len(base_tops)
    assert sorted(tops) == sorted(base_tops)


@patch("salt.utils.thin.gen_thin", MagicMock(return_value="/path/to/thin/thin.tgz"))
@patch("salt.utils.hashutils.get_hash", MagicMock(return_value=12345))
def test_thin_sum(thin_ctx):
    """
    Test thin.thin_sum function.

    :return:
    """
    assert thin.thin_sum("/cachedir", form="sha256")[1] == 12345
    thin.salt.utils.hashutils.get_hash.assert_called()
    assert thin.salt.utils.hashutils.get_hash.call_count == 1

    path, form = thin.salt.utils.hashutils.get_hash.call_args[0]
    assert path == "/path/to/thin/thin.tgz"
    assert form == "sha256"


@patch("salt.utils.thin.gen_min", MagicMock(return_value="/path/to/thin/min.tgz"))
@patch("salt.utils.hashutils.get_hash", MagicMock(return_value=12345))
def test_min_sum(thin_ctx):
    """
    Test thin.thin_sum function.

    :return:
    """
    assert thin.min_sum("/cachedir", form="sha256") == 12345
    thin.salt.utils.hashutils.get_hash.assert_called()
    assert thin.salt.utils.hashutils.get_hash.call_count == 1

    path, form = thin.salt.utils.hashutils.get_hash.call_args[0]
    assert path == "/path/to/thin/min.tgz"
    assert form == "sha256"


@patch("salt.utils.thin.sys.version_info", (2, 5))
@patch("salt.exceptions.SaltSystemExit", Exception)
def test_gen_thin_fails_ancient_python_version(thin_ctx):
    """
    Test thin.gen_thin function raises an exception
    if Python major/minor version is lower than 2.6

    :return:
    """
    with pytest.raises(salt.exceptions.SaltSystemExit) as err:
        thin.sys.exc_clear = lambda: None
        thin.gen_thin("")
    assert 'The minimum required python version to run salt-ssh is "3"' in str(
        err.value
    )


@patch("salt.exceptions.SaltSystemExit", Exception)
@patch("salt.utils.thin.log", MagicMock())
@patch("salt.utils.thin.os.makedirs", MagicMock())
@patch("salt.utils.files.fopen", MagicMock())
@patch("salt.utils.thin._get_salt_call", MagicMock())
@patch("salt.utils.thin._get_ext_namespaces", MagicMock())
@patch("salt.utils.thin.get_tops", MagicMock(return_value=["/foo3", "/bar3"]))
@patch("salt.utils.thin.get_ext_tops", MagicMock(return_value={}))
@patch("salt.utils.thin.os.path.isfile", MagicMock())
@patch("salt.utils.thin.os.path.isdir", MagicMock(return_value=True))
@patch("salt.utils.thin.log", MagicMock())
@patch("salt.utils.thin.os.remove", MagicMock())
@patch("salt.utils.thin.os.path.exists", MagicMock())
@patch("salt.utils.path.os_walk", MagicMock(return_value=[]))
@patch(
    "salt.utils.thin.subprocess.Popen",
    _popen(
        None,
        side_effect=[(bts("2.7"), bts("")), (bts('["/foo27", "/bar27"]'), bts(""))],
    ),
)
@patch("salt.utils.thin.tarfile", MagicMock())
@patch("salt.utils.thin.zipfile", MagicMock())
@patch("salt.utils.thin.os.getcwd", MagicMock())
@patch("salt.utils.thin.os.access", MagicMock(return_value=True))
@patch("salt.utils.thin.os.chdir", MagicMock())
@patch("salt.utils.thin.os.close", MagicMock())
@patch("salt.utils.thin.tempfile.mkdtemp", MagicMock())
@patch("salt.utils.thin.tempfile.mkstemp", MagicMock(return_value=(3, ".temporary")))
@patch("salt.utils.thin.shutil", MagicMock())
@patch("salt.utils.thin.sys.version_info", _version_info(3, 6))
@patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/python"))
def test_gen_thin_compression_fallback_py3(thin_ctx):
    """
    Test thin.gen_thin function if fallbacks to the gzip compression, once setup wrong.
    NOTE: Py2 version of this test is not required, as code shares the same spot across the versions.

    :return:
    """
    thin.gen_thin("", compress="arj")
    thin.log.warning.assert_called()
    pt, msg = thin.log.warning.mock_calls[0][1]
    assert (
        pt % msg
        == 'Unknown compression type: "arj". Falling back to "gzip" compression.'
    )
    thin.zipfile.ZipFile.assert_not_called()
    thin.tarfile.open.assert_called()


@patch("salt.exceptions.SaltSystemExit", Exception)
@patch("salt.utils.thin.log", MagicMock())
@patch("salt.utils.thin.os.makedirs", MagicMock())
@patch("salt.utils.files.fopen", MagicMock())
@patch("salt.utils.thin._get_salt_call", MagicMock())
@patch("salt.utils.thin._get_ext_namespaces", MagicMock())
@patch("salt.utils.thin.get_tops", MagicMock(return_value=["/foo3", "/bar3"]))
@patch("salt.utils.thin.get_ext_tops", MagicMock(return_value={}))
@patch("salt.utils.thin.os.path.isfile", MagicMock())
@patch("salt.utils.thin.os.path.isdir", MagicMock(return_value=False))
@patch("salt.utils.thin.log", MagicMock())
@patch("salt.utils.thin.os.remove", MagicMock())
@patch("salt.utils.thin.os.path.exists", MagicMock())
@patch("salt.utils.path.os_walk", MagicMock(return_value=[]))
@patch(
    "salt.utils.thin.subprocess.Popen",
    _popen(
        None,
        side_effect=[(bts("2.7"), bts("")), (bts('["/foo27", "/bar27"]'), bts(""))],
    ),
)
@patch("salt.utils.thin.tarfile", MagicMock())
@patch("salt.utils.thin.zipfile", MagicMock())
@patch("salt.utils.thin.os.getcwd", MagicMock())
@patch("salt.utils.thin.os.access", MagicMock(return_value=True))
@patch("salt.utils.thin.os.chdir", MagicMock())
@patch("salt.utils.thin.os.close", MagicMock())
@patch("salt.utils.thin.tempfile.mkdtemp", MagicMock(return_value=""))
@patch("salt.utils.thin.tempfile.mkstemp", MagicMock(return_value=(3, ".temporary")))
@patch("salt.utils.thin.shutil", MagicMock())
@patch("salt.utils.thin.sys.version_info", _version_info(3, 6))
@patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/python"))
def test_gen_thin_control_files_written_py3(thin_ctx):
    """
    Test thin.gen_thin function if control files are written (version, salt-call etc).
    :return:
    """
    thin.gen_thin("")
    arc_name, arc_mode = thin.tarfile.method_calls[0][1]
    assert arc_name == ".temporary"
    assert arc_mode == "w:gz"
    for idx, fname in enumerate(
        ["version", ".thin-gen-py-version", "salt-call", "supported-versions"]
    ):
        name = thin.tarfile.open().method_calls[idx + 2][1][0]
        assert name == fname
    thin.tarfile.open().close.assert_called()


@patch("salt.exceptions.SaltSystemExit", Exception)
@patch("salt.utils.thin.log", MagicMock())
@patch("salt.utils.thin.os.makedirs", MagicMock())
@patch("salt.utils.files.fopen", MagicMock())
@patch("salt.utils.thin._get_salt_call", MagicMock())
@patch("salt.utils.thin._get_ext_namespaces", MagicMock())
@patch("salt.utils.thin.get_tops", MagicMock(return_value=["/salt", "/bar3"]))
@patch("salt.utils.thin.get_ext_tops", MagicMock(return_value={}))
@patch("salt.utils.thin.os.path.isfile", MagicMock())
@patch("salt.utils.thin.os.path.isdir", MagicMock(return_value=True))
@patch("salt.utils.thin.log", MagicMock())
@patch("salt.utils.thin.os.remove", MagicMock())
@patch("salt.utils.thin.os.path.exists", MagicMock())
@patch(
    "salt.utils.path.os_walk",
    MagicMock(
        return_value=(
            ("root", [], ["r1", "r2", "r3"]),
            ("root2", [], ["r4", "r5", "r6"]),
        )
    ),
)
@patch("salt.utils.thin.tarfile", _tarfile(None))
@patch("salt.utils.thin.zipfile", MagicMock())
@patch(
    "salt.utils.thin.os.getcwd",
    MagicMock(return_value=os.path.join(RUNTIME_VARS.TMP, "fake-cwd")),
)
@patch("salt.utils.thin.os.chdir", MagicMock())
@patch("salt.utils.thin.os.close", MagicMock())
@patch("salt.utils.thin.tempfile.mkdtemp", MagicMock(return_value=""))
@patch("salt.utils.thin.tempfile.mkstemp", MagicMock(return_value=(3, ".temporary")))
@patch("salt.utils.thin.shutil", MagicMock())
@patch("salt.utils.thin.sys.version_info", _version_info(3, 6))
@patch("salt.utils.hashutils.DigestCollector", MagicMock())
@patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/python"))
def test_gen_thin_main_content_files_written_py3(thin_ctx):
    """
    Test thin.gen_thin function if main content files are written.
    NOTE: Py2 version of this test is not required, as code shares the same spot across the versions.

    :return:
    """
    thin.gen_thin("")
    files = []
    for py in ("py3", "pyall"):
        for i in range(1, 4):
            files.append(os.path.join(py, "root", f"r{i}"))
        for i in range(4, 7):
            files.append(os.path.join(py, "root2", f"r{i}"))
    for cl in thin.tarfile.open().method_calls[:-6]:
        arcname = cl[2].get("arcname")
        assert arcname in files
        files.pop(files.index(arcname))
    assert not files


@patch("salt.exceptions.SaltSystemExit", Exception)
@patch("salt.utils.thin.log", MagicMock())
@patch("salt.utils.thin.os.makedirs", MagicMock())
@patch("salt.utils.files.fopen", MagicMock())
@patch("salt.utils.thin._get_salt_call", MagicMock())
@patch("salt.utils.thin._get_ext_namespaces", MagicMock())
@patch("salt.utils.thin.get_tops", MagicMock(return_value=[]))
@patch(
    "salt.utils.thin.get_ext_tops",
    MagicMock(
        return_value={
            "namespace": {
                "py-version": [3, 0],
                "path": "/opt/2015.8/salt",
                "dependencies": ["/opt/certifi", "/opt/whatever"],
            }
        }
    ),
)
@patch("salt.utils.thin.os.path.isfile", MagicMock())
@patch("salt.utils.thin.os.path.isdir", MagicMock(return_value=True))
@patch("salt.utils.thin.log", MagicMock())
@patch("salt.utils.thin.os.remove", MagicMock())
@patch("salt.utils.thin.os.path.exists", MagicMock())
@patch(
    "salt.utils.path.os_walk",
    MagicMock(
        return_value=(
            ("root", [], ["r1", "r2", "r3"]),
            ("root2", [], ["r4", "r5", "r6"]),
        )
    ),
)
@patch("salt.utils.thin.tarfile", _tarfile(None))
@patch("salt.utils.thin.zipfile", MagicMock())
@patch(
    "salt.utils.thin.os.getcwd",
    MagicMock(return_value=os.path.join(RUNTIME_VARS.TMP, "fake-cwd")),
)
@patch("salt.utils.thin.os.chdir", MagicMock())
@patch("salt.utils.thin.os.close", MagicMock())
@patch("salt.utils.thin.tempfile.mkdtemp", MagicMock(return_value=""))
@patch("salt.utils.thin.tempfile.mkstemp", MagicMock(return_value=(3, ".temporary")))
@patch("salt.utils.thin.shutil", MagicMock())
@patch("salt.utils.thin.sys.version_info", _version_info(3, 6))
@patch("salt.utils.hashutils.DigestCollector", MagicMock())
@patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/python"))
def test_gen_thin_ext_alternative_content_files_written_py3(thin_ctx):
    """
    Test thin.gen_thin function if external alternative content files are written.
    :return:
    """
    ext_conf = {
        "namespace": {
            "py-version": [3, 0],
            "path": "/opt/2015.8/salt",
            "dependencies": {
                "certifi": "/opt/certifi",
                "whatever": "/opt/whatever",
            },
        }
    }

    thin.gen_thin("", extended_cfg=ext_conf)
    files = []
    for py in ("pyall", "pyall", "py3"):
        for i in range(1, 4):
            files.append(os.path.join("namespace", py, "root", f"r{i}"))
        for i in range(4, 7):
            files.append(os.path.join("namespace", py, "root2", f"r{i}"))

    for idx, cl in enumerate(thin.tarfile.open().method_calls[:-6]):
        arcname = cl[2].get("arcname")
        assert arcname in files
        files.pop(files.index(arcname))
    assert not files


def test_get_supported_py_config_typecheck(thin_ctx):
    """
    Test collecting proper py-versions. Should return bytes type.
    :return:
    """
    tops = {}
    ext_cfg = {}
    out = thin._get_supported_py_config(tops=tops, extended_cfg=ext_cfg)
    assert type(salt.utils.stringutils.to_bytes("")) == type(out)


def test_get_supported_py_config_base_tops(thin_ctx):
    """
    Test collecting proper py-versions. Should return proper base tops.
    :return:
    """
    tops = {"3": ["/groundkeepers", "/stole"], "2": ["/the-root", "/password"]}
    ext_cfg = {}
    out = (
        salt.utils.stringutils.to_str(
            thin._get_supported_py_config(tops=tops, extended_cfg=ext_cfg)
        )
        .strip()
        .split(os.linesep)
    )
    assert len(out) == 2
    for t_line in ["py3:3:0", "py2:2:7"]:
        assert t_line in out


def test_get_supported_py_config_ext_tops(thin_ctx):
    """
    Test collecting proper py-versions. Should return proper ext conf tops.
    :return:
    """
    tops = {}
    ext_cfg = {
        "solar-interference": {"py-version": [2, 6]},
        "second-system-effect": {"py-version": [2, 7]},
    }
    out = (
        salt.utils.stringutils.to_str(
            thin._get_supported_py_config(tops=tops, extended_cfg=ext_cfg)
        )
        .strip()
        .split(os.linesep)
    )
    for t_line in ["second-system-effect:2:7", "solar-interference:2:6"]:
        assert t_line in out


@patch("salt.exceptions.SaltSystemExit", Exception)
@patch("salt.utils.thin.log", MagicMock())
@patch("salt.utils.thin.os.makedirs", MagicMock())
@patch("salt.utils.files.fopen", MagicMock())
@patch("salt.utils.thin._get_salt_call", MagicMock())
@patch("salt.utils.thin._get_ext_namespaces", MagicMock())
@patch("salt.utils.thin.get_tops", MagicMock(return_value=["/foo3", "/bar3"]))
@patch("salt.utils.thin.get_ext_tops", MagicMock(return_value={}))
@patch("salt.utils.thin.os.path.isfile", MagicMock())
@patch("salt.utils.thin.os.path.isdir", MagicMock(return_value=False))
@patch("salt.utils.thin.log", MagicMock())
@patch("salt.utils.thin.os.remove", MagicMock())
@patch("salt.utils.thin.os.path.exists", MagicMock())
@patch("salt.utils.path.os_walk", MagicMock(return_value=[]))
@patch(
    "salt.utils.thin.subprocess.Popen",
    _popen(
        None,
        side_effect=[(bts("2.7"), bts("")), (bts('["/foo27", "/bar27"]'), bts(""))],
    ),
)
@patch("salt.utils.thin.tarfile", MagicMock())
@patch("salt.utils.thin.zipfile", MagicMock())
@patch("salt.utils.thin.os.getcwd", MagicMock())
@patch("salt.utils.thin.os.access", MagicMock(return_value=False))
@patch("salt.utils.thin.os.chdir", MagicMock())
@patch("salt.utils.thin.os.close", MagicMock())
@patch("salt.utils.thin.tempfile.mkdtemp", MagicMock(return_value=""))
@patch("salt.utils.thin.tempfile.mkstemp", MagicMock(return_value=(3, ".temporary")))
@patch("salt.utils.thin.shutil", MagicMock())
@patch("salt.utils.thin.sys.version_info", _version_info(3, 6))
def test_gen_thin_control_files_written_access_denied_cwd(thin_ctx):
    """
    Test thin.gen_thin function if control files are written (version, salt-call etc)
    when the current working directory is inaccessible, eg. Salt is configured to run as
    a non-root user but the command is executed in a directory that the user does not
    have permissions to.  Issue #54317.

    NOTE: Py2 version of this test is not required, as code shares the same spot across the versions.

    :return:
    """
    thin.gen_thin("")
    arc_name, arc_mode = thin.tarfile.method_calls[0][1]
    assert arc_name == ".temporary"
    assert arc_mode == "w:gz"
    for idx, fname in enumerate(
        ["version", ".thin-gen-py-version", "salt-call", "supported-versions"]
    ):
        name = thin.tarfile.open().method_calls[idx + 2][1][0]
        assert name == fname
    thin.tarfile.open().close.assert_called()


def test_get_tops_python(thin_ctx):
    """
    test get_tops_python
    """
    patch_proc = patch(
        "salt.utils.thin.subprocess.Popen",
        _popen(
            None,
            side_effect=[
                (bts("jinja2/__init__.py"), bts("")),
                (bts("yaml/__init__.py"), bts("")),
                (bts("tornado/__init__.py"), bts("")),
                (bts("msgpack/__init__.py"), bts("")),
                (bts("certifi/__init__.py"), bts("")),
                (bts("singledispatch.py"), bts("")),
                (bts(""), bts("")),
                (bts(""), bts("")),
                (bts(""), bts("")),
                (bts(""), bts("")),
                (bts(""), bts("")),
                (bts("looseversion.py"), bts("")),
                (bts("packaging/__init__.py"), bts("")),
                (bts("backports/__init__.py"), bts("")),
                (bts("distro.py"), bts("")),
            ],
        ),
    )

    patch_os = patch("os.path.exists", return_value=True)
    patch_which = patch("salt.utils.path.which", return_value=True)
    with patch_proc, patch_os, patch_which:
        with TstSuiteLoggingHandler() as log_handler:
            exp_ret = copy.deepcopy(thin_ctx.exp_ret)
            ret = thin.get_tops_python("python3.7", ext_py_ver=[3, 7])
            if salt.utils.platform.is_windows():
                for key, value in ret.items():
                    ret[key] = str(pathlib.Path(value).resolve(strict=False))
                for key, value in exp_ret.items():
                    exp_ret[key] = str(pathlib.Path(value).resolve(strict=False))
            assert ret == exp_ret
            assert (
                "ERROR:Could not auto detect file location for module concurrent"
                " for python version python3.7" in log_handler.messages
            )


def test_get_tops_python_exclude(thin_ctx):
    """
    test get_tops_python when excluding modules
    """
    patch_proc = patch(
        "salt.utils.thin.subprocess.Popen",
        _popen(
            None,
            side_effect=[
                (bts("tornado/__init__.py"), bts("")),
                (bts("msgpack/__init__.py"), bts("")),
                (bts("certifi/__init__.py"), bts("")),
                (bts("singledispatch.py"), bts("")),
                (bts(""), bts("")),
                (bts(""), bts("")),
                (bts(""), bts("")),
                (bts(""), bts("")),
                (bts(""), bts("")),
                (bts("looseversion.py"), bts("")),
                (bts("packaging/__init__.py"), bts("")),
                (bts("backports/__init__.py"), bts("")),
                (bts("distro.py"), bts("")),
            ],
        ),
    )
    exp_ret = copy.deepcopy(thin_ctx.exp_ret)
    for lib in thin_ctx.exc_libs:
        exp_ret.pop(lib)

    patch_os = patch("os.path.exists", return_value=True)
    patch_which = patch("salt.utils.path.which", return_value=True)
    with patch_proc, patch_os, patch_which:
        ret = thin.get_tops_python(
            "python3.7", exclude=thin_ctx.exc_libs, ext_py_ver=[3, 7]
        )
        if salt.utils.platform.is_windows():
            for key, value in ret.items():
                ret[key] = str(pathlib.Path(value).resolve(strict=False))
            for key, value in exp_ret.items():
                exp_ret[key] = str(pathlib.Path(value).resolve(strict=False))
        assert ret == exp_ret


def test_pack_alternatives_exclude(thin_ctx):
    """
    test pack_alternatives when mixing
    manually set dependencies and auto
    detecting other modules.
    """
    patch_proc = patch(
        "salt.utils.thin.subprocess.Popen",
        _popen(
            None,
            side_effect=[
                (bts(thin_ctx.fake_libs["distro"]), bts("")),
                (bts(thin_ctx.fake_libs["yaml"]), bts("")),
                (bts(thin_ctx.fake_libs["tornado"]), bts("")),
                (bts(thin_ctx.fake_libs["msgpack"]), bts("")),
                (bts(""), bts("")),
                (bts(""), bts("")),
                (bts(""), bts("")),
                (bts(""), bts("")),
                (bts(""), bts("")),
                (bts(""), bts("")),
                (bts(""), bts("")),
                (bts("looseversion.py"), bts("")),
                (bts("packaging/__init__.py"), bts("")),
            ],
        ),
    )

    patch_os = patch("os.path.exists", return_value=True)
    ext_conf = copy.deepcopy(thin_ctx.ext_conf)
    ext_conf["test"]["auto_detect"] = True

    for lib in thin_ctx.fake_libs.values():
        os.makedirs(lib)
        with salt.utils.files.fopen(os.path.join(lib, "__init__.py"), "w+") as fp_:
            fp_.write("test")

    exp_files = thin_ctx.exp_files.copy()
    exp_files.extend(
        [
            os.path.join("yaml", "__init__.py"),
            os.path.join("tornado", "__init__.py"),
            os.path.join("msgpack", "__init__.py"),
        ]
    )

    patch_which = patch("salt.utils.path.which", return_value=True)

    with patch_os, patch_proc, patch_which:
        thin._pack_alternative(ext_conf, thin_ctx.digest, thin_ctx.tar)
        calls = thin_ctx.tar.mock_calls
        for _file in exp_files:
            assert [x for x in calls if f"{_file}" in x[-2]]


def test_pack_alternatives(thin_ctx):
    """
    test thin._pack_alternatives
    """
    with patch("salt.utils.thin.get_ext_tops", MagicMock(return_value=thin_ctx.tops)):
        thin._pack_alternative(thin_ctx.ext_conf, thin_ctx.digest, thin_ctx.tar)
        calls = thin_ctx.tar.mock_calls
        for _file in thin_ctx.exp_files:
            assert [x for x in calls if f"{_file}" in x[-2]]
            assert [
                x
                for x in calls
                if os.path.join("test", "pyall", _file) in x[-1]["arcname"]
            ]


def test_pack_alternatives_not_normalized(thin_ctx):
    """
    test thin._pack_alternatives when the path
    is not normalized
    """
    tops = copy.deepcopy(thin_ctx.tops)
    tops["test"]["dependencies"] = [thin_ctx.jinja_fp + "/"]
    with patch("salt.utils.thin.get_ext_tops", MagicMock(return_value=tops)):
        thin._pack_alternative(thin_ctx.ext_conf, thin_ctx.digest, thin_ctx.tar)
        calls = thin_ctx.tar.mock_calls
        for _file in thin_ctx.exp_files:
            assert [x for x in calls if f"{_file}" in x[-2]]
            assert [
                x
                for x in calls
                if os.path.join("test", "pyall", _file) in x[-1]["arcname"]
            ]


def test_pack_alternatives_path_doesnot_exist(thin_ctx):
    """
    test thin._pack_alternatives when the path
    doesnt exist. Check error log message
    and expect that because the directory
    does not exist jinja2 does not get
    added to the tar
    """
    bad_path = os.path.join(tempfile.gettempdir(), "doesnotexisthere")
    tops = copy.deepcopy(thin_ctx.tops)
    tops["test"]["dependencies"] = [bad_path]
    with patch("salt.utils.thin.get_ext_tops", MagicMock(return_value=tops)):
        with TstSuiteLoggingHandler() as log_handler:
            thin._pack_alternative(thin_ctx.ext_conf, thin_ctx.digest, thin_ctx.tar)
            msg = "ERROR:File path {} does not exist. Unable to add to salt-ssh thin".format(
                bad_path
            )
            assert msg in log_handler.messages
    calls = thin_ctx.tar.mock_calls
    for _file in thin_ctx.exp_files:
        arg = [x for x in calls if f"{_file}" in x[-2]]
        kwargs = [
            x for x in calls if os.path.join("test", "pyall", _file) in x[-1]["arcname"]
        ]
        if "jinja2" in _file:
            assert not arg
            assert not kwargs
        else:
            assert arg
            assert kwargs


def test_pack_alternatives_auto_detect(thin_ctx):
    """
    test thin._pack_alternatives when auto_detect
    is enabled
    """
    ext_conf = copy.deepcopy(thin_ctx.ext_conf)
    ext_conf["test"]["auto_detect"] = True

    for lib in thin_ctx.fake_libs.values():
        os.makedirs(lib)
        with salt.utils.files.fopen(os.path.join(lib, "__init__.py"), "w+") as fp_:
            fp_.write("test")

    patch_tops_py = patch(
        "salt.utils.thin.get_tops_python", return_value=thin_ctx.fake_libs
    )

    exp_files = thin_ctx.exp_files.copy()
    exp_files.extend(
        [
            os.path.join("yaml", "__init__.py"),
            os.path.join("tornado", "__init__.py"),
            os.path.join("msgpack", "__init__.py"),
        ]
    )
    with patch_tops_py:
        thin._pack_alternative(ext_conf, thin_ctx.digest, thin_ctx.tar)
        calls = thin_ctx.tar.mock_calls
        for _file in exp_files:
            assert [x for x in calls if f"{_file}" in x[-2]]


def test_pack_alternatives_empty_dependencies(thin_ctx):
    """
    test _pack_alternatives when dependencies is not
    set in the config.
    """
    ext_conf = copy.deepcopy(thin_ctx.ext_conf)
    ext_conf["test"]["auto_detect"] = True
    ext_conf["test"].pop("dependencies")

    for lib in thin_ctx.fake_libs.values():
        os.makedirs(lib)
        with salt.utils.files.fopen(os.path.join(lib, "__init__.py"), "w+") as fp_:
            fp_.write("test")

    patch_tops_py = patch(
        "salt.utils.thin.get_tops_python", return_value=thin_ctx.fake_libs
    )

    exp_files = thin_ctx.exp_files.copy()
    exp_files.extend(
        [
            os.path.join("yaml", "__init__.py"),
            os.path.join("tornado", "__init__.py"),
            os.path.join("msgpack", "__init__.py"),
        ]
    )
    with patch_tops_py:
        thin._pack_alternative(ext_conf, thin_ctx.digest, thin_ctx.tar)
        calls = thin_ctx.tar.mock_calls
        for _file in exp_files:
            assert [x for x in calls if f"{_file}" in x[-2]]


@pytest.mark.slow_test
@pytest.mark.skip_on_windows(reason="salt-ssh does not deploy to/from windows")
def test_thin_dir(thin_ctx):
    """
    Test the thin dir to make sure salt-call can run

    Run salt call via a python in a new virtual environment to ensure
    salt-call has all dependencies needed.
    """
    # This was previously an integration test and is now here, as a unit test.
    # Should actually be a functional test
    orig_which = shutil.which
    with patch(
        "tests.support.helpers.shutil.which",
        side_effect=lambda cmd, *args, **kwargs: (
            sys.executable if cmd == "python" else orig_which(cmd, *args, **kwargs)
        ),
    ):
        with VirtualEnv() as venv:
            salt.utils.thin.gen_thin(str(venv.venv_dir))
            thin_dir = venv.venv_dir / "thin"
            thin_archive = thin_dir / "thin.tgz"
            tar = tarfile.open(str(thin_archive))
            tar.extractall(str(thin_dir))  # nosec
            tar.close()
            python_bin = pathlib.Path(venv.venv_python)
            if not python_bin.exists():
                for candidate in (
                    "python3",
                    f"python{sys.version_info[0]}",
                    f"python{sys.version_info[0]}.{sys.version_info[1]}",
                ):
                    alt = python_bin.parent / candidate
                    if alt.exists():
                        python_bin = alt
                        break
            ret = venv.run(
                str(python_bin),
                str(thin_dir / "salt-call"),
                "--version",
                check=False,
            )
        assert ret.returncode == 0, ret



"""
Utils for proxy.
"""

import logging

import salt.utils.platform

log = logging.getLogger(__file__)


def is_proxytype(opts, proxytype):
    """
    Is this a proxy minion of type proxytype
    """
    return (
        salt.utils.platform.is_proxy()
        and opts.get("proxy", {}).get("proxytype", None) == proxytype
    )



"""
URL utils
"""

import re
import sys
from urllib.parse import urlparse, urlunparse

import salt.utils.data
import salt.utils.path
import salt.utils.platform
import salt.utils.versions


def parse(url):
    """
    Parse a salt:// URL; return the path and a possible saltenv query.
    """
    if not url.startswith("salt://"):
        return url, None

    # urlparse will split on valid filename chars such as '?' and '&'
    resource = url.split("salt://", 1)[-1]

    if "?env=" in resource:
        # "env" is not supported; Use "saltenv".
        path, saltenv = resource.split("?env=", 1)[0], None
    elif "?saltenv=" in resource:
        path, saltenv = resource.split("?saltenv=", 1)
    else:
        path, saltenv = resource, None

    if salt.utils.platform.is_windows():
        path = salt.utils.path.sanitize_win_path(path)

    return path, saltenv


def create(path, saltenv=None):
    """
    join `path` and `saltenv` into a 'salt://' URL.
    """
    path = path.replace("\\", "/")
    if salt.utils.platform.is_windows():
        path = salt.utils.path.sanitize_win_path(path)
    path = salt.utils.data.decode(path)

    query = f"saltenv={saltenv}" if saltenv else ""
    url = salt.utils.data.decode(urlunparse(("file", "", path, "", query, "")))
    return "salt://{}".format(url[len("file:///") :])


def is_escaped(url):
    """
    test whether `url` is escaped with `|`
    """
    scheme = urlparse(url).scheme
    if not scheme:
        return url.startswith("|")
    elif scheme == "salt":
        path, saltenv = parse(url)
        if salt.utils.platform.is_windows() and "|" in url:
            return path.startswith("_")
        else:
            return path.startswith("|")
    else:
        return False


def escape(url):
    """
    add escape character `|` to `url`
    """
    if salt.utils.platform.is_windows():
        return url

    scheme = urlparse(url).scheme
    if not scheme:
        if url.startswith("|"):
            return url
        else:
            return f"|{url}"
    elif scheme == "salt":
        path, saltenv = parse(url)
        if path.startswith("|"):
            return create(path, saltenv)
        else:
            return create(f"|{path}", saltenv)
    else:
        return url


def unescape(url):
    """
    remove escape character `|` from `url`
    """
    scheme = urlparse(url).scheme
    if not scheme:
        return url.lstrip("|")
    elif scheme == "salt":
        path, saltenv = parse(url)
        if salt.utils.platform.is_windows() and "|" in url:
            return create(path.lstrip("_"), saltenv)
        else:
            return create(path.lstrip("|"), saltenv)
    else:
        return url


def add_env(url, saltenv):
    """
    append `saltenv` to `url` as a query parameter to a 'salt://' url
    """
    if not url.startswith("salt://"):
        return url

    path, senv = parse(url)
    return create(path, saltenv)


def split_env(url):
    """
    remove the saltenv query parameter from a 'salt://' url
    """
    if not url.startswith("salt://"):
        return url, None

    path, senv = parse(url)
    return create(path), senv


def validate(url, protos):
    """
    Return true if the passed URL scheme is in the list of accepted protos
    """
    if urlparse(url).scheme in protos:
        return True
    return False


def strip_proto(url):
    """
    Return a copy of the string with the protocol designation stripped, if one
    was present.
    """
    return re.sub("^[^:/]+://", "", url)


def add_http_basic_auth(url, user=None, password=None, https_only=False):
    """
    Return a string with http basic auth incorporated into it
    """
    if user is None and password is None:
        return url
    else:
        urltuple = urlparse(url)
        if https_only and urltuple.scheme != "https":
            raise ValueError("Basic Auth only supported for HTTPS")
        if password is None:
            netloc = f"{user}@{urltuple.netloc}"
            urltuple = urltuple._replace(netloc=netloc)
            return urlunparse(urltuple)
        else:
            netloc = f"{user}:{password}@{urltuple.netloc}"
            urltuple = urltuple._replace(netloc=netloc)
            return urlunparse(urltuple)


def redact_http_basic_auth(output):
    """
    Remove HTTP user and password
    """
    # We can't use re.compile because re.compile(someregex).sub() doesn't
    # support flags even in Python 2.7.
    url_re = "(https?)://.*@"
    redacted = r"\1://<redacted>@"
    if sys.version_info >= (2, 7):
        # re.sub() supports flags as of 2.7, use this to do a case-insensitive
        # match.
        return re.sub(url_re, redacted, output, flags=re.IGNORECASE)
    else:
        # We're on python 2.6, test if a lowercased version of the output
        # string matches the regex...
        if re.search(url_re, output.lower()):
            # ... and if it does, perform the regex substitution.
            return re.sub(url_re, redacted, output.lower())
    # No match, just return the original string
    return output



"""
This module allows SPM to use the local filesystem to install files for SPM.

.. versionadded:: 2015.8.0
"""

import logging
import os
import os.path

import salt.syspaths
import salt.utils.files
import salt.utils.stringutils

# Get logging started
log = logging.getLogger(__name__)
FILE_TYPES = ("c", "d", "g", "l", "r", "s", "m")
# c: config file
# d: documentation file
# g: ghost file (i.e. the file contents are not included in the package payload)
# l: license file
# r: readme file
# s: SLS file
# m: Salt module


def init(**kwargs):
    """
    Initialize the directories for the files
    """
    formula_path = __opts__["formula_path"]
    pillar_path = __opts__["pillar_path"]
    reactor_path = __opts__["reactor_path"]
    for dir_ in (formula_path, pillar_path, reactor_path):
        if not os.path.exists(dir_):
            os.makedirs(dir_)
    return {
        "formula_path": formula_path,
        "pillar_path": pillar_path,
        "reactor_path": reactor_path,
    }


def check_existing(package, pkg_files, formula_def, conn=None):
    """
    Check the filesystem for existing files
    """
    if conn is None:
        conn = init()

    node_type = str(__opts__.get("spm_node_type"))

    existing_files = []
    for member in pkg_files:
        if member.isdir():
            continue

        tld = formula_def.get("top_level_dir", package)
        new_name = member.name.replace(f"{package}/", "")
        if not new_name.startswith(tld):
            continue

        if member.name.startswith(f"{package}/_"):
            if node_type in ("master", "minion"):
                # Module files are distributed via extmods directory
                out_file = os.path.join(
                    salt.syspaths.CACHE_DIR,
                    node_type,
                    "extmods",
                    new_name.replace("_", ""),
                )
            else:
                # Module files are distributed via _modules, _states, etc
                out_file = os.path.join(conn["formula_path"], new_name)
        elif member.name == f"{package}/pillar.example":
            # Pillars are automatically put in the pillar_path
            new_name = f"{package}.sls.orig"
            out_file = os.path.join(conn["pillar_path"], new_name)
        elif package.endswith("-conf"):
            # Configuration files go into /etc/salt/
            out_file = os.path.join(salt.syspaths.CONFIG_DIR, new_name)
        elif package.endswith("-reactor"):
            # Reactor files go into /srv/reactor/
            out_file = os.path.join(conn["reactor_path"], member.name)
        else:
            out_file = os.path.join(conn["formula_path"], member.name)

        if os.path.exists(out_file):
            existing_files.append(out_file)
            if not __opts__["force"]:
                log.error("%s already exists, not installing", out_file)

    return existing_files


def install_file(package, formula_tar, member, formula_def, conn=None):
    """
    Install a single file to the file system
    """
    if member.name == package:
        return False

    if conn is None:
        conn = init()

    node_type = str(__opts__.get("spm_node_type"))

    out_path = conn["formula_path"]

    tld = formula_def.get("top_level_dir", package)
    new_name = member.name.replace(f"{package}/", "", 1)
    if (
        not new_name.startswith(tld)
        and not new_name.startswith("_")
        and not new_name.startswith("pillar.example")
        and not new_name.startswith("README")
    ):
        log.debug("%s not in top level directory, not installing", new_name)
        return False

    for line in formula_def.get("files", []):
        tag = ""
        for ftype in FILE_TYPES:
            if line.startswith(f"{ftype}|"):
                tag = line.split("|", 1)[0]
                line = line.split("|", 1)[1]
        if tag and new_name == line:
            if tag in ("c", "d", "g", "l", "r"):
                out_path = __opts__["spm_share_dir"]
            elif tag in ("s", "m"):
                pass

    if member.name.startswith(f"{package}{os.sep}_"):
        if node_type in ("master", "minion"):
            # Module files are distributed via extmods directory
            member.name = new_name.replace(f"{package}{os.sep}_", "")
            out_path = os.path.join(
                salt.syspaths.CACHE_DIR,
                node_type,
                "extmods",
            )
        else:
            # Module files are distributed via _modules, _states, etc
            member.name = new_name.replace(f"{package}{os.sep}", "")
    elif member.name == f"{package}/pillar.example":
        # Pillars are automatically put in the pillar_path
        member.name = f"{package}.sls.orig"
        out_path = conn["pillar_path"]
    elif package.endswith("-conf"):
        # Configuration files go into /etc/salt/
        member.name = member.name.replace(f"{package}/", "")
        out_path = salt.syspaths.CONFIG_DIR
    elif package.endswith("-reactor"):
        # Reactor files go into /srv/reactor/
        out_path = __opts__["reactor_path"]

    # This ensures that double directories (i.e., apache/apache/) don't
    # get created
    comps = member.path.split("/")
    if len(comps) > 1 and comps[0] == comps[1]:
        member.path = "/".join(comps[1:])

    log.debug("Installing package file %s to %s", member.name, out_path)
    formula_tar.extract(member, out_path)

    return out_path


def remove_file(path, conn=None):
    """
    Remove a single file from the file system
    """
    if conn is None:
        conn = init()

    log.debug("Removing package file %s", path)
    os.remove(path)


def hash_file(path, hashobj, conn=None):
    """
    Get the hexdigest hash value of a file
    """
    if os.path.isdir(path):
        return ""

    with salt.utils.files.fopen(path, "r") as f:
        hashobj.update(salt.utils.stringutils.to_bytes(f.read()))
        return hashobj.hexdigest()


def path_exists(path):
    """
    Check to see whether the file already exists
    """
    return os.path.exists(path)


def path_isdir(path):
    """
    Check to see whether the file already exists
    """
    return os.path.isdir(path)



#
#   Proxy minion metaproxy modules
#
import concurrent.futures
import copy
import logging
import os
import signal
import threading
import traceback
import types

import salt
import salt._logging
import salt.beacons
import salt.cli.daemons
import salt.client
import salt.config
import salt.crypt
import salt.defaults.exitcodes
import salt.engines
import salt.ext.tornado.gen  # pylint: disable=F0401
import salt.ext.tornado.ioloop  # pylint: disable=F0401
import salt.loader
import salt.minion
import salt.payload
import salt.pillar
import salt.serializers.msgpack
import salt.syspaths
import salt.utils.args
import salt.utils.context
import salt.utils.data
import salt.utils.dictupdate
import salt.utils.error
import salt.utils.event
import salt.utils.files
import salt.utils.jid
import salt.utils.minion
import salt.utils.minions
import salt.utils.network
import salt.utils.platform
import salt.utils.process
import salt.utils.schedule
import salt.utils.ssdp
import salt.utils.user
import salt.utils.zeromq
from salt.defaults import DEFAULT_TARGET_DELIM
from salt.exceptions import (
    CommandExecutionError,
    CommandNotFoundError,
    SaltInvocationError,
    SaltSystemExit,
)
from salt.minion import ProxyMinion
from salt.utils.event import tagify
from salt.utils.process import SignalHandlingProcess, default_signals

log = logging.getLogger(__name__)


def post_master_init(self, master):
    """
    Function to finish init after a deltaproxy proxy
    minion has finished connecting to a master.

    This is primarily loading modules, pillars, etc. (since they need
    to know which master they connected to)
    """

    if self.connected:
        self.opts["pillar"] = yield salt.pillar.get_async_pillar(
            self.opts,
            self.opts["grains"],
            self.opts["id"],
            saltenv=self.opts["saltenv"],
            pillarenv=self.opts.get("pillarenv"),
        ).compile_pillar()

        # Ensure that the value of master is the one we passed in.
        # if pillar_opts is enabled then master could be overwritten
        # when compile_pillar is run.
        self.opts["master"] = master

        tag = "salt/deltaproxy/start"
        self._fire_master(tag=tag)

    if "proxy" not in self.opts["pillar"] and "proxy" not in self.opts:
        errmsg = (
            "No proxy key found in pillar or opts for id {}. Check your pillar/opts "
            "configuration and contents.  Salt-proxy aborted.".format(self.opts["id"])
        )
        log.error(errmsg)
        self._running = False
        raise SaltSystemExit(code=-1, msg=errmsg)

    if "proxy" not in self.opts:
        self.opts["proxy"] = self.opts["pillar"]["proxy"]

    pillar = copy.deepcopy(self.opts["pillar"])
    pillar.pop("master", None)
    self.opts = salt.utils.dictupdate.merge(
        self.opts,
        pillar,
        strategy=self.opts.get("proxy_merge_pillar_in_opts_strategy"),
        merge_lists=self.opts.get("proxy_deep_merge_pillar_in_opts", False),
    )

    if self.opts.get("proxy_mines_pillar"):
        # Even when not required, some details such as mine configuration
        # should be merged anyway whenever possible.
        if "mine_interval" in self.opts["pillar"]:
            self.opts["mine_interval"] = self.opts["pillar"]["mine_interval"]
        if "mine_functions" in self.opts["pillar"]:
            general_proxy_mines = self.opts.get("mine_functions", [])
            specific_proxy_mines = self.opts["pillar"]["mine_functions"]
            try:
                self.opts["mine_functions"] = general_proxy_mines + specific_proxy_mines
            except TypeError as terr:
                log.error(
                    "Unable to merge mine functions from the pillar in the opts, for proxy %s",
                    self.opts["id"],
                )

    fq_proxyname = self.opts["proxy"]["proxytype"]

    # Need to load the modules so they get all the dunder variables
    (
        self.functions,
        self.returners,
        self.function_errors,
        self.executors,
    ) = self._load_modules()

    # we can then sync any proxymodules down from the master
    # we do a sync_all here in case proxy code was installed by
    # SPM or was manually placed in /srv/salt/_modules etc.
    self.functions["saltutil.sync_all"](saltenv=self.opts["saltenv"])

    # Pull in the utils
    self.utils = salt.loader.utils(self.opts)

    # Then load the proxy module
    self.proxy = salt.loader.proxy(self.opts, utils=self.utils)

    # And re-load the modules so the __proxy__ variable gets injected
    (
        self.functions,
        self.returners,
        self.function_errors,
        self.executors,
    ) = self._load_modules()
    self.functions.pack["__proxy__"] = self.proxy
    self.proxy.pack["__salt__"] = self.functions
    self.proxy.pack["__ret__"] = self.returners
    self.proxy.pack["__pillar__"] = self.opts["pillar"]

    # Reload utils as well (chicken and egg, __utils__ needs __proxy__ and __proxy__ needs __utils__
    self.utils = salt.loader.utils(self.opts, proxy=self.proxy)
    self.proxy.pack["__utils__"] = self.utils

    # Reload all modules so all dunder variables are injected
    self.proxy.reload_modules()

    # Start engines here instead of in the Minion superclass __init__
    # This is because we need to inject the __proxy__ variable but
    # it is not setup until now.
    self.io_loop.spawn_callback(
        salt.engines.start_engines, self.opts, self.process_manager, proxy=self.proxy
    )

    proxy_init_func_name = f"{fq_proxyname}.init"
    proxy_shutdown_func_name = f"{fq_proxyname}.shutdown"
    if (
        proxy_init_func_name not in self.proxy
        or proxy_shutdown_func_name not in self.proxy
    ):
        errmsg = (
            "Proxymodule {} is missing an init() or a shutdown() or both. "
            "Check your proxymodule.  Salt-proxy aborted.".format(fq_proxyname)
        )
        log.error(errmsg)
        self._running = False
        raise SaltSystemExit(code=-1, msg=errmsg)

    self.module_executors = self.proxy.get(
        f"{fq_proxyname}.module_executors", lambda: []
    )()
    proxy_init_fn = self.proxy[proxy_init_func_name]
    proxy_init_fn(self.opts)

    self.opts["grains"] = salt.loader.grains(self.opts, proxy=self.proxy)

    self.mod_opts = self._prep_mod_opts()
    self.matchers = salt.loader.matchers(self.opts)
    self.beacons = salt.beacons.Beacon(self.opts, self.functions)
    uid = salt.utils.user.get_uid(user=self.opts.get("user", None))
    self.proc_dir = salt.minion.get_proc_dir(self.opts["cachedir"], uid=uid)

    if self.connected and self.opts["pillar"]:
        # The pillar has changed due to the connection to the master.
        # Reload the functions so that they can use the new pillar data.
        (
            self.functions,
            self.returners,
            self.function_errors,
            self.executors,
        ) = self._load_modules()
        if hasattr(self, "schedule"):
            self.schedule.functions = self.functions
            self.schedule.returners = self.returners

    if not hasattr(self, "schedule"):
        self.schedule = salt.utils.schedule.Schedule(
            self.opts,
            self.functions,
            self.returners,
            cleanup=[salt.minion.master_event(type="alive")],
            proxy=self.proxy,
            _subprocess_list=self.subprocess_list,
        )

    # add default scheduling jobs to the minions scheduler
    if self.opts["mine_enabled"] and "mine.update" in self.functions:
        self.schedule.add_job(
            {
                "__mine_interval": {
                    "function": "mine.update",
                    "minutes": self.opts["mine_interval"],
                    "jid_include": True,
                    "maxrunning": 2,
                    "run_on_start": True,
                    "return_job": self.opts.get("mine_return_job", False),
                }
            },
            persist=True,
            fire_event=False,
        )
        log.info("Added mine.update to scheduler")
    else:
        self.schedule.delete_job("__mine_interval", persist=True, fire_event=False)

    # add master_alive job if enabled
    if self.opts["transport"] != "tcp" and self.opts["master_alive_interval"] > 0:
        self.schedule.add_job(
            {
                salt.minion.master_event(type="alive", master=self.opts["master"]): {
                    "function": "status.master",
                    "seconds": self.opts["master_alive_interval"],
                    "jid_include": True,
                    "maxrunning": 1,
                    "return_job": False,
                    "kwargs": {"master": self.opts["master"], "connected": True},
                }
            },
            persist=True,
            fire_event=False,
        )
        if (
            self.opts["master_failback"]
            and "master_list" in self.opts
            and self.opts["master"] != self.opts["master_list"][0]
        ):
            self.schedule.add_job(
                {
                    salt.minion.master_event(type="failback"): {
                        "function": "status.ping_master",
                        "seconds": self.opts["master_failback_interval"],
                        "jid_include": True,
                        "maxrunning": 1,
                        "return_job": False,
                        "kwargs": {"master": self.opts["master_list"][0]},
                    }
                },
                persist=True,
                fire_event=False,
            )
        else:
            self.schedule.delete_job(
                salt.minion.master_event(type="failback"),
                persist=True,
                fire_event=False,
            )
    else:
        self.schedule.delete_job(
            salt.minion.master_event(type="alive", master=self.opts["master"]),
            persist=True,
            fire_event=False,
        )
        self.schedule.delete_job(
            salt.minion.master_event(type="failback"),
            persist=True,
            fire_event=False,
        )

    # proxy keepalive
    proxy_alive_fn = fq_proxyname + ".alive"
    if (
        proxy_alive_fn in self.proxy
        and "status.proxy_reconnect" in self.functions
        and self.opts.get("proxy_keep_alive", True)
    ):
        # if `proxy_keep_alive` is either not specified, either set to False does not retry reconnecting
        self.schedule.add_job(
            {
                "__proxy_keepalive": {
                    "function": "status.proxy_reconnect",
                    "minutes": self.opts.get(
                        "proxy_keep_alive_interval", 1
                    ),  # by default, check once per minute
                    "jid_include": True,
                    "maxrunning": 1,
                    "return_job": False,
                    "kwargs": {"proxy_name": fq_proxyname},
                }
            },
            persist=True,
            fire_event=False,
        )
        self.schedule.enable_schedule(fire_event=False)
    else:
        self.schedule.delete_job(
            "__proxy_keepalive",
            persist=True,
            fire_event=False,
        )

    #  Sync the grains here so the proxy can communicate them to the master
    self.functions["saltutil.sync_grains"](saltenv="base")
    self.grains_cache = self.opts["grains"]
    # Now setup the deltaproxies
    self.deltaproxy = {}
    self.deltaproxy_opts = {}
    self.deltaproxy_objs = {}
    self.proxy_grains = {}
    self.proxy_pillar = {}
    self.proxy_context = {}
    self.add_periodic_callback("cleanup", self.cleanup_subprocesses)

    _failed = list()
    if self.opts["proxy"].get("parallel_startup"):
        log.debug("Initiating parallel startup for proxies")
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = {
                executor.submit(
                    subproxy_post_master_init,
                    _id,
                    uid,
                    self.opts,
                    self.proxy,
                    self.utils,
                ): _id
                for _id in self.opts["proxy"].get("ids", [])
            }

        for future in concurrent.futures.as_completed(futures):
            try:
                sub_proxy_data = future.result()
            except Exception as exc:  # pylint: disable=broad-except
                _id = futures[future]
                log.info(
                    "An exception occured during initialization for %s, skipping: %s",
                    _id,
                    exc,
                )
                _failed.append(_id)
                continue
            minion_id = sub_proxy_data["proxy_opts"].get("id")

            if sub_proxy_data["proxy_minion"]:
                self.deltaproxy_opts[minion_id] = sub_proxy_data["proxy_opts"]
                self.deltaproxy_objs[minion_id] = sub_proxy_data["proxy_minion"]

                if self.deltaproxy_opts[minion_id] and self.deltaproxy_objs[minion_id]:
                    self.deltaproxy_objs[minion_id].req_channel = (
                        salt.channel.client.AsyncReqChannel.factory(
                            sub_proxy_data["proxy_opts"], io_loop=self.io_loop
                        )
                    )
    else:
        log.debug("Initiating non-parallel startup for proxies")
        for _id in self.opts["proxy"].get("ids", []):
            try:
                sub_proxy_data = subproxy_post_master_init(
                    _id, uid, self.opts, self.proxy, self.utils
                )
            except Exception as exc:  # pylint: disable=broad-except
                log.info(
                    "An exception occured during initialization for %s, skipping: %s",
                    _id,
                    exc,
                )
                _failed.append(_id)
                continue
            minion_id = sub_proxy_data["proxy_opts"].get("id")

            if sub_proxy_data["proxy_minion"]:
                self.deltaproxy_opts[minion_id] = sub_proxy_data["proxy_opts"]
                self.deltaproxy_objs[minion_id] = sub_proxy_data["proxy_minion"]

                if self.deltaproxy_opts[minion_id] and self.deltaproxy_objs[minion_id]:
                    self.deltaproxy_objs[minion_id].req_channel = (
                        salt.channel.client.AsyncReqChannel.factory(
                            sub_proxy_data["proxy_opts"], io_loop=self.io_loop
                        )
                    )

    if _failed:
        log.info("Following sub proxies failed %s", _failed)
    self.ready = True


def subproxy_post_master_init(minion_id, uid, opts, main_proxy, main_utils):
    """
    Function to finish init after a deltaproxy proxy
    minion has finished connecting to a master.

    This is primarily loading modules, pillars, etc. (since they need
    to know which master they connected to) for the sub proxy minions.
    """
    proxy_grains = {}
    proxy_pillar = {}

    proxyopts = opts.copy()
    proxyopts["id"] = minion_id

    proxyopts = salt.config.proxy_config(
        opts["conf_file"], defaults=proxyopts, minion_id=minion_id
    )
    proxyopts.update({"id": minion_id, "proxyid": minion_id, "subproxy": True})

    proxy_context = {"proxy_id": minion_id}

    # We need grains first to be able to load pillar, which is where we keep the proxy
    # configurations
    proxy_grains = salt.loader.grains(
        proxyopts, proxy=main_proxy, context=proxy_context
    )
    proxy_pillar = salt.pillar.get_pillar(
        proxyopts,
        proxy_grains,
        minion_id,
        saltenv=proxyopts["saltenv"],
        pillarenv=proxyopts.get("pillarenv"),
    ).compile_pillar()

    proxyopts["proxy"] = proxy_pillar.get("proxy", {})
    if not proxyopts["proxy"]:
        log.warning(
            "Pillar data for proxy minion %s could not be loaded, skipping.", minion_id
        )
        return {"proxy_minion": None, "proxy_opts": {}}

    # Remove ids
    proxyopts["proxy"].pop("ids", None)

    proxyopts.update(
        {
            "pillar": proxy_pillar,
            "grains": proxy_grains,
            "hash_id": opts["id"],
        }
    )

    _proxy_minion = ProxyMinion(proxyopts)
    _proxy_minion.proc_dir = salt.minion.get_proc_dir(proxyopts["cachedir"], uid=uid)

    # And load the modules
    (
        _proxy_minion.functions,
        _proxy_minion.returners,
        _proxy_minion.function_errors,
        _proxy_minion.executors,
    ) = _proxy_minion._load_modules(
        opts=proxyopts,
        grains=proxyopts["grains"],
        context=proxy_context,
    )

    # we can then sync any proxymodules down from the master
    # we do a sync_all here in case proxy code was installed by
    # SPM or was manually placed in /srv/salt/_modules etc.
    _proxy_minion.functions["saltutil.sync_all"](saltenv=opts["saltenv"])

    # And re-load the modules so the __proxy__ variable gets injected
    (
        _proxy_minion.functions,
        _proxy_minion.returners,
        _proxy_minion.function_errors,
        _proxy_minion.executors,
    ) = _proxy_minion._load_modules(
        opts=proxyopts,
        grains=proxyopts["grains"],
        context=proxy_context,
    )

    # Create this after modules are synced to ensure
    # any custom modules, eg. custom proxy modules
    # are avaiable.
    _proxy_minion.proxy = salt.loader.proxy(
        proxyopts, utils=main_utils, context=proxy_context
    )

    _proxy_minion.functions.pack["__proxy__"] = _proxy_minion.proxy
    _proxy_minion.proxy.pack["__salt__"] = _proxy_minion.functions
    _proxy_minion.proxy.pack["__ret__"] = _proxy_minion.returners
    _proxy_minion.proxy.pack["__pillar__"] = proxyopts["pillar"]
    _proxy_minion.proxy.pack["__grains__"] = proxyopts["grains"]

    # Reload utils as well (chicken and egg, __utils__ needs __proxy__ and __proxy__ needs __utils__
    _proxy_minion.proxy.utils = salt.loader.utils(
        proxyopts, proxy=_proxy_minion.proxy, context=proxy_context
    )

    _proxy_minion.proxy.pack["__utils__"] = _proxy_minion.proxy.utils

    # Reload all modules so all dunder variables are injected
    _proxy_minion.proxy.reload_modules()

    _proxy_minion.connected = True

    _fq_proxyname = proxyopts["proxy"]["proxytype"]

    proxy_init_fn = _proxy_minion.proxy[_fq_proxyname + ".init"]
    try:
        proxy_init_fn(proxyopts)
    except Exception as exc:  # pylint: disable=broad-except
        log.error(
            "An exception occured during the initialization of minion %s: %s",
            minion_id,
            exc,
            exc_info=True,
        )
        return {"proxy_minion": None, "proxy_opts": {}}

    # Reload the grains
    proxy_grains = salt.loader.grains(
        proxyopts, proxy=_proxy_minion.proxy, context=proxy_context
    )
    proxyopts["grains"] = proxy_grains

    if not hasattr(_proxy_minion, "schedule"):
        _proxy_minion.schedule = salt.utils.schedule.Schedule(
            proxyopts,
            _proxy_minion.functions,
            _proxy_minion.returners,
            cleanup=[salt.minion.master_event(type="alive")],
            proxy=_proxy_minion.proxy,
            new_instance=True,
            _subprocess_list=_proxy_minion.subprocess_list,
        )

    # proxy keepalive
    _proxy_alive_fn = _fq_proxyname + ".alive"
    if (
        _proxy_alive_fn in _proxy_minion.proxy
        and "status.proxy_reconnect" in _proxy_minion.functions
        and proxyopts.get("proxy_keep_alive", True)
    ):
        # if `proxy_keep_alive` is either not specified, either set to False does not retry reconnecting
        _proxy_minion.schedule.add_job(
            {
                "__proxy_keepalive": {
                    "function": "status.proxy_reconnect",
                    "minutes": proxyopts.get(
                        "proxy_keep_alive_interval", 1
                    ),  # by default, check once per minute
                    "jid_include": True,
                    "maxrunning": 1,
                    "return_job": False,
                    "kwargs": {"proxy_name": _fq_proxyname},
                }
            },
            persist=True,
            fire_event=False,
        )
        _proxy_minion.schedule.enable_schedule(fire_event=False)
    else:
        _proxy_minion.schedule.delete_job(
            "__proxy_keepalive", persist=True, fire_event=False
        )

    return {"proxy_minion": _proxy_minion, "proxy_opts": proxyopts}


def target(cls, minion_instance, opts, data, connected, creds_map):
    """
    Handle targeting of the minion.

    Calling _thread_multi_return or _thread_return
    depending on a single or multiple commands.
    """
    log.debug(
        "Deltaproxy minion_instance %s(ID: %s). Target: %s",
        minion_instance,
        minion_instance.opts["id"],
        opts["id"],
    )
    if creds_map:
        salt.crypt.AsyncAuth.creds_map = creds_map

    if not hasattr(minion_instance, "proc_dir"):
        uid = salt.utils.user.get_uid(user=opts.get("user", None))
        minion_instance.proc_dir = salt.minion.get_proc_dir(opts["cachedir"], uid=uid)

    with salt.ext.tornado.stack_context.StackContext(minion_instance.ctx):
        if isinstance(data["fun"], tuple) or isinstance(data["fun"], list):
            ProxyMinion._thread_multi_return(minion_instance, opts, data)
        else:
            ProxyMinion._thread_return(minion_instance, opts, data)


def thread_return(cls, minion_instance, opts, data):
    """
    This method should be used as a threading target, start the actual
    minion side execution.
    """
    fn_ = os.path.join(minion_instance.proc_dir, data["jid"])

    if opts["multiprocessing"] and not salt.utils.platform.spawning_platform():

        # Shutdown the multiprocessing before daemonizing
        salt._logging.shutdown_logging()

        salt.utils.process.daemonize_if(opts)

        # Reconfigure multiprocessing logging after daemonizing
        salt._logging.setup_logging()

    salt.utils.process.appendproctitle(f"{cls.__name__}._thread_return")

    sdata = {"pid": os.getpid()}
    sdata.update(data)
    log.info("Starting a new job with PID %s", sdata["pid"])
    with salt.utils.files.fopen(fn_, "w+b") as fp_:
        fp_.write(salt.payload.dumps(sdata))
    ret = {"success": False}
    function_name = data["fun"]
    executors = (
        data.get("module_executors")
        or getattr(minion_instance, "module_executors", [])
        or opts.get("module_executors", ["direct_call"])
    )
    allow_missing_funcs = any(
        [
            minion_instance.executors[f"{executor}.allow_missing_func"](function_name)
            for executor in executors
            if f"{executor}.allow_missing_func" in minion_instance.executors
        ]
    )
    if function_name in minion_instance.functions or allow_missing_funcs is True:
        try:
            minion_blackout_violation = False
            if minion_instance.connected and minion_instance.opts["pillar"].get(
                "minion_blackout", False
            ):
                whitelist = minion_instance.opts["pillar"].get(
                    "minion_blackout_whitelist", []
                )
                # this minion is blacked out. Only allow saltutil.refresh_pillar and the whitelist
                if (
                    function_name != "saltutil.refresh_pillar"
                    and function_name not in whitelist
                ):
                    minion_blackout_violation = True
            # use minion_blackout_whitelist from grains if it exists
            if minion_instance.opts["grains"].get("minion_blackout", False):
                whitelist = minion_instance.opts["grains"].get(
                    "minion_blackout_whitelist", []
                )
                if (
                    function_name != "saltutil.refresh_pillar"
                    and function_name not in whitelist
                ):
                    minion_blackout_violation = True
            if minion_blackout_violation:
                raise SaltInvocationError(
                    'Minion in blackout mode. Set "minion_blackout" '
                    "to False in pillar or grains to resume operations. Only "
                    "saltutil.refresh_pillar allowed in blackout mode."
                )

            if function_name in minion_instance.functions:
                func = minion_instance.functions[function_name]
                args, kwargs = salt.minion.load_args_and_kwargs(func, data["arg"], data)
            else:
                # only run if function_name is not in minion_instance.functions and allow_missing_funcs is True
                func = function_name
                args, kwargs = data["arg"], data
            minion_instance.functions.pack["__context__"]["retcode"] = 0
            minion_instance.functions.pack["__opts__"] = opts
            if isinstance(executors, str):
                executors = [executors]
            elif not isinstance(executors, list) or not executors:
                raise SaltInvocationError(
                    "Wrong executors specification: {}. String or non-empty list expected".format(
                        executors
                    )
                )
            if opts.get("sudo_user", "") and executors[-1] != "sudo":
                executors[-1] = "sudo"  # replace the last one with sudo
            log.debug("Executors list %s", executors)

            for name in executors:
                fname = f"{name}.execute"
                if fname not in minion_instance.executors:
                    raise SaltInvocationError(f"Executor '{name}' is not available")

                return_data = minion_instance.executors[fname](
                    opts, data, func, args, kwargs
                )
                if return_data is not None:
                    break

            if isinstance(return_data, types.GeneratorType):
                ind = 0
                iret = {}
                for single in return_data:
                    if isinstance(single, dict) and isinstance(iret, dict):
                        iret.update(single)
                    else:
                        if not iret:
                            iret = []
                        iret.append(single)
                    tag = tagify([data["jid"], "prog", opts["id"], str(ind)], "job")
                    event_data = {"return": single}
                    minion_instance._fire_master(event_data, tag)
                    ind += 1
                ret["return"] = iret
            else:
                ret["return"] = return_data

            retcode = minion_instance.functions.pack["__context__"].get(
                "retcode", salt.defaults.exitcodes.EX_OK
            )
            if retcode == salt.defaults.exitcodes.EX_OK:
                # No nonzero retcode in __context__ dunder. Check if return
                # is a dictionary with a "result" or "success" key.
                try:
                    func_result = all(
                        return_data.get(x, True) for x in ("result", "success")
                    )
                except Exception:  # pylint: disable=broad-except
                    # return data is not a dict
                    func_result = True
                if not func_result:
                    retcode = salt.defaults.exitcodes.EX_GENERIC

            ret["retcode"] = retcode
            ret["success"] = retcode == salt.defaults.exitcodes.EX_OK
        except CommandNotFoundError as exc:
            msg = f'Command required for "{function_name}" not found'
            log.debug(msg, exc_info=True)
            ret["return"] = f"{msg}: {exc}"
            ret["out"] = "nested"
            ret["retcode"] = salt.defaults.exitcodes.EX_GENERIC
        except CommandExecutionError as exc:
            log.error(
                'A command in "%s" had a problem: %s',
                function_name,
                exc,
                exc_info_on_loglevel=logging.DEBUG,
            )
            ret["return"] = f"ERROR: {exc}"
            ret["out"] = "nested"
            ret["retcode"] = salt.defaults.exitcodes.EX_GENERIC
        except SaltInvocationError as exc:
            log.error(
                'Problem executing "%s": %s',
                function_name,
                exc,
                exc_info_on_loglevel=logging.DEBUG,
            )
            ret["return"] = f'ERROR executing "{function_name}": {exc}'
            ret["out"] = "nested"
            ret["retcode"] = salt.defaults.exitcodes.EX_GENERIC
        except TypeError as exc:
            msg = "Passed invalid arguments to {}: {}\n{}".format(
                function_name, exc, func.__doc__ or ""
            )
            log.warning(msg, exc_info_on_loglevel=logging.DEBUG)
            ret["return"] = msg
            ret["out"] = "nested"
            ret["retcode"] = salt.defaults.exitcodes.EX_GENERIC
        except Exception:  # pylint: disable=broad-except
            msg = "The minion function caused an exception"
            log.warning(msg, exc_info=True)
            salt.utils.error.fire_exception(
                salt.exceptions.MinionError(msg), opts, job=data
            )
            ret["return"] = f"{msg}: {traceback.format_exc()}"
            ret["out"] = "nested"
            ret["retcode"] = salt.defaults.exitcodes.EX_GENERIC
    else:
        docs = minion_instance.functions["sys.doc"](f"{function_name}*")
        if docs:
            docs[function_name] = minion_instance.functions.missing_fun_string(
                function_name
            )
            ret["return"] = docs
        else:
            ret["return"] = minion_instance.functions.missing_fun_string(function_name)
            mod_name = function_name.split(".")[0]
            if mod_name in minion_instance.function_errors:
                ret["return"] += ' Possible reasons: "{}"'.format(
                    minion_instance.function_errors[mod_name]
                )
        ret["success"] = False
        ret["retcode"] = salt.defaults.exitcodes.EX_GENERIC
        ret["out"] = "nested"

    ret["jid"] = data["jid"]
    ret["fun"] = data["fun"]
    ret["fun_args"] = data["arg"]
    if "master_id" in data:
        ret["master_id"] = data["master_id"]
    if "metadata" in data:
        if isinstance(data["metadata"], dict):
            ret["metadata"] = data["metadata"]
        else:
            log.warning("The metadata parameter must be a dictionary. Ignoring.")
    if minion_instance.connected:
        minion_instance._return_pub(ret, timeout=minion_instance._return_retry_timer())

    # Add default returners from minion config
    # Should have been coverted to comma-delimited string already
    if isinstance(opts.get("return"), str):
        if data["ret"]:
            data["ret"] = ",".join((data["ret"], opts["return"]))
        else:
            data["ret"] = opts["return"]

    # TODO: make a list? Seems odd to split it this late :/
    if data["ret"] and isinstance(data["ret"], str):
        if "ret_config" in data:
            ret["ret_config"] = data["ret_config"]
        if "ret_kwargs" in data:
            ret["ret_kwargs"] = data["ret_kwargs"]
        ret["id"] = opts["id"]
        for returner in set(data["ret"].split(",")):
            try:
                returner_str = f"{returner}.returner"
                if returner_str in minion_instance.returners:
                    minion_instance.returners[returner_str](ret)
                else:
                    returner_err = minion_instance.returners.missing_fun_string(
                        returner_str
                    )
                    log.error(
                        "Returner %s could not be loaded: %s",
                        returner_str,
                        returner_err,
                    )
            except Exception as exc:  # pylint: disable=broad-except
                log.exception("The return failed for job %s: %s", data["jid"], exc)


def thread_multi_return(cls, minion_instance, opts, data):
    """
    This method should be used as a threading target, start the actual
    minion side execution.
    """
    fn_ = os.path.join(minion_instance.proc_dir, data["jid"])

    if opts["multiprocessing"] and not salt.utils.platform.spawning_platform():
        # Shutdown the multiprocessing before daemonizing
        salt._logging.shutdown_logging()

        salt.utils.process.daemonize_if(opts)

        # Reconfigure multiprocessing logging after daemonizing
        salt._logging.setup_logging()

    salt.utils.process.appendproctitle(f"{cls.__name__}._thread_multi_return")

    sdata = {"pid": os.getpid()}
    sdata.update(data)
    log.info("Starting a new job with PID %s", sdata["pid"])
    with salt.utils.files.fopen(fn_, "w+b") as fp_:
        fp_.write(salt.payload.dumps(sdata))

    multifunc_ordered = opts.get("multifunc_ordered", False)
    num_funcs = len(data["fun"])
    if multifunc_ordered:
        ret = {
            "return": [None] * num_funcs,
            "retcode": [None] * num_funcs,
            "success": [False] * num_funcs,
        }
    else:
        ret = {"return": {}, "retcode": {}, "success": {}}

    for ind in range(0, num_funcs):
        if not multifunc_ordered:
            ret["success"][data["fun"][ind]] = False
        try:
            minion_blackout_violation = False
            if minion_instance.connected and minion_instance.opts["pillar"].get(
                "minion_blackout", False
            ):
                whitelist = minion_instance.opts["pillar"].get(
                    "minion_blackout_whitelist", []
                )
                # this minion is blacked out. Only allow saltutil.refresh_pillar and the whitelist
                if (
                    data["fun"][ind] != "saltutil.refresh_pillar"
                    and data["fun"][ind] not in whitelist
                ):
                    minion_blackout_violation = True
            elif minion_instance.opts["grains"].get("minion_blackout", False):
                whitelist = minion_instance.opts["grains"].get(
                    "minion_blackout_whitelist", []
                )
                if (
                    data["fun"][ind] != "saltutil.refresh_pillar"
                    and data["fun"][ind] not in whitelist
                ):
                    minion_blackout_violation = True
            if minion_blackout_violation:
                raise SaltInvocationError(
                    'Minion in blackout mode. Set "minion_blackout" '
                    "to False in pillar or grains to resume operations. Only "
                    "saltutil.refresh_pillar allowed in blackout mode."
                )

            func = minion_instance.functions[data["fun"][ind]]

            args, kwargs = salt.minion.load_args_and_kwargs(
                func, data["arg"][ind], data
            )
            minion_instance.functions.pack["__context__"]["retcode"] = 0
            key = ind if multifunc_ordered else data["fun"][ind]
            ret["return"][key] = func(*args, **kwargs)
            retcode = minion_instance.functions.pack["__context__"].get("retcode", 0)
            if retcode == 0:
                # No nonzero retcode in __context__ dunder. Check if return
                # is a dictionary with a "result" or "success" key.
                try:
                    func_result = all(
                        ret["return"][key].get(x, True) for x in ("result", "success")
                    )
                except Exception:  # pylint: disable=broad-except
                    # return data is not a dict
                    func_result = True
                if not func_result:
                    retcode = 1

            ret["retcode"][key] = retcode
            ret["success"][key] = retcode == 0
        except Exception as exc:  # pylint: disable=broad-except
            trb = traceback.format_exc()
            log.warning("The minion function caused an exception: %s", exc)
            if multifunc_ordered:
                ret["return"][ind] = trb
            else:
                ret["return"][data["fun"][ind]] = trb
        ret["jid"] = data["jid"]
        ret["fun"] = data["fun"]
        ret["fun_args"] = data["arg"]
    if "metadata" in data:
        ret["metadata"] = data["metadata"]
    if minion_instance.connected:
        minion_instance._return_pub(ret, timeout=minion_instance._return_retry_timer())
    if data["ret"]:
        if "ret_config" in data:
            ret["ret_config"] = data["ret_config"]
        if "ret_kwargs" in data:
            ret["ret_kwargs"] = data["ret_kwargs"]
        for returner in set(data["ret"].split(",")):
            ret["id"] = opts["id"]
            try:
                minion_instance.returners[f"{returner}.returner"](ret)
            except Exception as exc:  # pylint: disable=broad-except
                log.error("The return failed for job %s: %s", data["jid"], exc)


def handle_payload(self, payload):
    """
    Verify the publication and then pass
    the payload along to _handle_decoded_payload.
    """
    if payload is not None and payload["enc"] == "aes":
        # First handle payload for the "control" proxy
        if self._target_load(payload["load"]):
            self._handle_decoded_payload(payload["load"])

        # The following handles the sub-proxies
        sub_ids = self.opts["proxy"].get("ids", [self.opts["id"]])
        for _id in sub_ids:
            if _id in self.deltaproxy_objs:
                instance = self.deltaproxy_objs[_id]
                if instance._target_load(payload["load"]):
                    instance._handle_decoded_payload(payload["load"])
            else:
                log.warning("Proxy minion %s is not loaded, skipping.", _id)

    elif self.opts["zmq_filtering"]:
        # In the filtering enabled case, we"d like to know when minion sees something it shouldnt
        log.trace(
            "Broadcast message received not for this minion, Load: %s", payload["load"]
        )
    # If it"s not AES, and thus has not been verified, we do nothing.
    # In the future, we could add support for some clearfuncs, but
    # the minion currently has no need.


def handle_decoded_payload(self, data):
    """
    Override this method if you wish to handle the decoded data
    differently.
    """
    if "user" in data:
        log.info(
            "User %s Executing command %s with jid %s",
            data["user"],
            data["fun"],
            data["jid"],
        )
    else:
        log.info("Executing command %s with jid %s", data["fun"], data["jid"])
    log.debug("Command details %s", data)

    # Don"t duplicate jobs
    log.trace("Started JIDs: %s", self.jid_queue)
    if self.jid_queue is not None:
        if data["jid"] in self.jid_queue:
            return
        else:
            self.jid_queue.append(data["jid"])
            if len(self.jid_queue) > self.opts["minion_jid_queue_hwm"]:
                self.jid_queue.pop(0)

    if isinstance(data["fun"], str):
        if data["fun"] == "sys.reload_modules":
            (
                self.functions,
                self.returners,
                self.function_errors,
                self.executors,
            ) = self._load_modules()
            self.schedule.functions = self.functions
            self.schedule.returners = self.returners

    process_count_max = self.opts.get("process_count_max")
    if process_count_max > 0:
        process_count = self.subprocess_list.count
        once_logged = False
        while process_count >= process_count_max:
            if once_logged is False:
                log.debug(
                    "Maximum number of processes reached while executing jid %s, waiting...",
                    data["jid"],
                )
                once_logged = True
            yield salt.ext.tornado.gen.sleep(0.5)
            process_count = self.subprocess_list.count

    # We stash an instance references to allow for the socket
    # communication in Windows. You can"t pickle functions, and thus
    # python needs to be able to reconstruct the reference on the other
    # side.
    instance = self
    multiprocessing_enabled = self.opts.get("multiprocessing", True)
    name = "ProcessPayload(jid={})".format(data["jid"])
    creds_map = None
    if multiprocessing_enabled:
        if salt.utils.platform.spawning_platform():
            # let python reconstruct the minion on the other side if we"re
            # running on spawning platforms
            instance = None
            creds_map = salt.crypt.AsyncAuth.creds_map
        with default_signals(signal.SIGINT, signal.SIGTERM):
            process = SignalHandlingProcess(
                target=target,
                args=(self, instance, self.opts, data, self.connected, creds_map),
                name=name,
            )
    else:
        process = threading.Thread(
            target=target,
            args=(self, instance, self.opts, data, self.connected, creds_map),
            name=name,
        )

    if multiprocessing_enabled:
        with default_signals(signal.SIGINT, signal.SIGTERM):
            # Reset current signals before starting the process in
            # order not to inherit the current signal handlers
            process.start()
    else:
        process.start()
    self.subprocess_list.add(process)


def target_load(self, load):
    """
    Verify that the publication is valid.
    """
    for key in ("tgt", "jid", "fun", "arg"):
        if key not in load:
            return False
    # Verify that the publication applies to this minion

    # It"s important to note that the master does some pre-processing
    # to determine which minions to send a request to. So for example,
    # a "salt -G "grain_key:grain_val" test.ping" will invoke some
    # pre-processing on the master and this minion should not see the
    # publication if the master does not determine that it should.

    if "tgt_type" in load:
        match_func = self.matchers.get("{}_match.match".format(load["tgt_type"]), None)
        if match_func is None:
            return False
        if load["tgt_type"] in ("grain", "grain_pcre", "pillar"):
            delimiter = load.get("delimiter", DEFAULT_TARGET_DELIM)
            if not match_func(load["tgt"], delimiter=delimiter, opts=self.opts):
                return False
        elif not match_func(load["tgt"], opts=self.opts):
            return False
    else:
        if not self.matchers["glob_match.match"](load["tgt"], opts=self.opts):
            return False

    return True


# Main Minion Tune In
def tune_in(self, start=True):
    """
    Lock onto the publisher. This is the main event loop for the minion
    :rtype : None
    """
    if self.opts["proxy"].get("parallel_startup"):
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(subproxy_tune_in, self.deltaproxy_objs[proxy_minion])
                for proxy_minion in self.deltaproxy_objs
            ]

        for f in concurrent.futures.as_completed(futures):
            _proxy_minion = f.result()
            log.debug("Tune in for sub proxy %r finished", _proxy_minion.opts.get("id"))
    else:
        for proxy_minion in self.deltaproxy_objs:
            _proxy_minion = subproxy_tune_in(self.deltaproxy_objs[proxy_minion])
            log.debug("Tune in for sub proxy %r finished", _proxy_minion.opts.get("id"))
    super(ProxyMinion, self).tune_in(start=start)


def subproxy_tune_in(proxy_minion, start=True):
    """
    Tunein sub proxy minions
    """
    proxy_minion.setup_scheduler()
    proxy_minion.setup_beacons()
    proxy_minion.add_periodic_callback("cleanup", proxy_minion.cleanup_subprocesses)
    proxy_minion._state_run()

    return proxy_minion
